from __future__ import print_function, division
import numpy as np
import sys, warnings
import time as systime
from copy import deepcopy

import gilly_utilities.system

# Compatability for python3
if getattr(sys.version_info, 'major') == 3:
    xrange = range

class array:
    """Fastest method of combining data for very large dataset. Much faster than
    np.append, or x.extend/append. Boo
    
    Example
    -------
    import Gilly_Utilities as gu
    
    #Define array
    myarray = gu.array()
    
    #Add data to array
    myarray.update([1,2,3,4,5])
    
    #To access data you need to finalise it
    myarray = myarray.finalise(dtype=int)
    
    myarray
    >>> [1,2,3,4,5]
        
    Source
    ------
    https://stackoverflow.com/questions/7133885/fastest-way-to-grow-a-numpy-numeric-array"""
    
    def __init__(self, data=None):
        self.data = [] if data is None else data
        self.flatten = lambda l: [item for sublist in l for item in sublist]
        
    def update(self, row):
        for r in row:
            self.data.append(r)

    def finalize(self, dtype=float, flat=True, dim='1D', method='fast', error_return=[]):
                
        if flat is True:
            if dim == '1D':
                if method is 'fast':
                    #Used to flatten out a NON NxM array which often occurs when you .update different sized arrays
                    #It first converts to an array (typical dtype is object) then uses hstack with the .flat command
                    #to completely flatten to a 1D array and then we convert to the correct dtype (if possible).
                    
                    #This method is typically the fastest way of flattening all the data. There are some issues that
                    #will impact the speed. Also, if no data exists when finalizing the array, an error will occur.
                    try:
                        return np.hstack(np.array(self.data).flat).astype(dtype)
                    except IndexError:
                        #If an error occurs in flattening data (e.g. self.data dimensions are 1 and cannot be combined)
                        print("Finalize Error")
                        method = 'ravel'
                if method is 'slow':
                    #Typically used as a last ditch attempt at flattening array when no other method works
                    #return np.array(self.flatten(self.data), dtype=dtype)
                    return flatten(self.data, type='ndarray', dtype=dtype)
                if method is 'object':
                    #Use in the special cases when using object arrays of different shapes (e.g. a.shape = (20,8), b.shape = (64,))
                    return np.hstack(np.array(self.flatten(self.data)).flat).astype(dtype)
                if method is 'ravel':
                    #A common method to flatten array which is very fast. Sometime faster than the hstack method. This
                    #is why we use this method when an error occurs in the hstack method.
                    return np.array(self.data, dtype=dtype).ravel()
                else:
                    sys.exit("[Gilly_Utilities] You must select the correct array method. Choose either: ['fast', 'slow', 'ravel']")
            elif dim == '2D':
                if len(self.data) == 0: return error_return
                
                if method is 'fast':
                    return np.hstack(self.data).astype(dtype)
                elif method is 'hstack':
                    return np.hstack(self.data).astype(dtype)
                elif method is 'vstack':
                    return np.vstack(self.data).astype(dtype)
                elif method is 'slow':
                    return self.flatten(self.data)
                
        else:
            return np.array(self.data, dtype=dtype)


class np_func(list):
    def ljust(self, n, fillvalue='', dtype=float):
        """Pads an array, list, by an integer number, n, with values
        set by the fillvalue. The output array type can also be
        set using dtype"""
        
        return np.array(self + [fillvalue] * (n - len(self)), dtype=dtype)
        
    def rjust(self, n, fillvalue='', dtype=float):
        """Pads an array, list, by an integer number, n, with values
        set by the fillvalue. The output array type can also be
        set using dtype"""
        
        return np.array([fillvalue] * (n - len(self)) + self, dtype=dtype)
    
    def rljust(self, n, fillvalue='', dtype=float):
        """Pads an array, list, by an integer number, n, with values
        set by the fillvalue. This will fill both left and right 
        sides of array"""
        
        ret = np.array([fillvalue] * np.int(np.ceil((n - len(self))/2)) + self + [fillvalue] * np.int(np.floor((n - len(self))/2)), dtype=dtype)
        
        if ret.size != n:
            return np.append(ret, fillvalue)
        else:
            return ret
        
class list_func(list):
    def ljust(self, n, fillvalue=''):
        """Pads an array, list, by an integer number, n, with values
        set by the fillvalue. The output array type can also be
        set using array_type"""
        
        return self + [fillvalue] * (n - len(self))
        
    def rjust(self, n, fillvalue='', array_type=int):
        """Pads an array, list, by an integer number, n, with values
        set by the fillvalue. The output array type can also be
        set using array_type"""
        
        return [fillvalue] * (n - len(self)) + self
    
    def rljust(self, n, fillvalue='', array_type=int):
        """Pads an array, list, by an integer number, n, with values
        set by the fillvalue. This will fill both left and right 
        sides of array"""
        
        return [fillvalue] * np.int(np.ceil((n - len(self))/2)) + self + [fillvalue] * np.int(np.floor((n - len(self))/2))


def subset(haystack, range, invert=False, impose=None):
    """Subsets an array, haystack, between the upper and lower values in range. There are parameters
    to invert the range selection (i.e. removes everything between range) and a parameter to use the
    information between haystack and range and place that over a new array in impose"""
    
    if invert is False:
        if impose is None:
            return haystack[(haystack >= range[0]) & (haystack < range[1])]
        else:
            if impose.size == haystack.size:
                return impose[(haystack >= range[0]) & (haystack < range[1])]
            else:
                sys.exit("[Error in subset] impose must be the same size as haystack")
                
    elif invert is True:
        if impose is None:
            return haystack[(haystack <= range[0]) | (haystack > range[1])]
        else:
            if impose.size == haystack.size:
                return impose[(haystack <= range[0]) | (haystack > range[1])]
            else:
                sys.exit("[Error in subset] impose must be the same size as haystack")
    
    else:
        sys.exit("[Error in subset] impose must be either None or an array of size haystack")    
        
def upset1d(haystack_large, haystack_small, sort=False, index=False):
    """Increase the size of haystack_small to match the dimensions of haystack_large. This is done
    by riffling the smaller array into the larger array. Only works if both input arrays are 1
    dimensional."""
    
    haystack_large = np.array(haystack_large) if sort is False else np.sort(np.array(haystack_large), kind='mergesort')
    haystack_small = np.array(haystack_small) if sort is False else np.sort(np.array(haystack_small), kind='mergesort')
    
    
    #Determine index locations
    idx = np.searchsorted(haystack_small, haystack_large)
    
    #Correct for index if out of bounds
    idx[idx == haystack_small.size] -= 1
    
    if index is False:
        return haystack_small[idx]    
    else:
        return idx

def upset2d(haystack_large, haystack_small, needles, sort=False):
    """Similar to upset1d but will work for 2d arrays. In this version you also need to specify needles.
    Needles is the dimensions of the large and small arrays which follows must be formatted into a loose
    2xM array where M is the number of dimensions.
    
    Example
    -------
    
    >>> haystack_large_time = np.arange(200)
    >>> haystack_large_height = np.arange(80)
    
    >>> haystack_small_time = np.arange(100)
    >>> haystack_small_height = np.arange(40)
    
    >>> needles = np.array([[haystack_large_time, haystack_large_height],[haystack_small_time, haystack_small_height]])
    
    >>> needles[0,0] == haystack_large_time
    True
    
    """
    
    masks = [upset1d(needles[0,i], needles[1,i], sort=sort, index=True) for i in xrange(needles.shape[1])]
    
    try:
        return haystack_small[masks[1]][:,masks[0]]    
    except:
        return haystack_small[masks[0]]

def midpoint(array):
    """Finds the mid point values of any array. The output will have a size, array.size + 1"""
    
    #Find difference between each element of array
    diff = np.diff(array)/2
    
    #Add difference between each element onto original array and taken into account boundaries
    #return np.hstack((array[0]-diff[0], array[1]-diff[1], array[1:] + diff))
    return np.hstack((array[0]-diff[0], array[:-1] + diff, array[-1] + diff[-1]))
        
def mask(haystack, needle, invert=False, impose=False, cross_impose=False, sort=False, find_all=False, approx=False):
    """Determines the all the needles in the haystack. Both haystack and needle
    need to be ndarrays for this too work. The advantage of this method is that 
    its FAST.
    
    Parameters
    ----------
    haystack : the array you want to search
    
    needle : the search values
    
    invert : invert the select indices in haystack array
    
    impose : returned is the haystack array with found indices masked out with value of impose
    
    cross_impose : if you select all values in haystack_a but you want to use those index locations
        to populate a different array, use cross_impose instead.
        
        Example:
        
        t = np.arange(10)
        z = np.array([-2,-1,0])
        
        bob = gu.mask(t, [1,5,9], invert=True, cross_impose=z)
        bob = [0,-2,2,3,4,-1,6,7,8,0]
        
    sort : specify if you require the data to be sorted first
    
    find_all : for each needle, if there is multiple solutions then find_all == False will only
        find the first solution as the haystack array is assumed to be sorted. if find_all == True
        then all solutions are selected.
    
    approx : Search for elements in haystack that are closest to needle elements. Similar to 
        gu.near and gu.argnear
        
    Reference
    ---------
    https://stackoverflow.com/a/31789252
    """
    
    haystack = np.array(haystack) if sort is False else np.sort(np.array(haystack), kind='mergesort')
    needle = np.array([needle])

    if approx is True: return searchsorted(haystack, needle)
    
    #Determine index locations
    idx = searchsorted(haystack, needle, side='left') if find_all is False else np.in1d(haystack, needle)
    
    if invert is False:
        if find_all is False:
            mask = idx < haystack.size
            mask[mask] = haystack[idx[mask]] == needle[mask]
            idx = idx[mask]
        else:
            mask = np.ones(haystack.size, dtype=np.bool)
            mask[idx] = 0
            idx = np.arange(haystack.size)[~mask]
    else:
        mask_invert = np.ones(haystack.size, dtype=np.bool)
        mask_invert[idx] = 0
        idx = np.arange(haystack.size)[mask_invert]

    if impose is not False and cross_impose is False:
        haystack_copy = haystack.copy()
        haystack_copy[idx] = impose
        
        return haystack_copy
    
    if cross_impose is not False:
        if np.array(cross_impose).size != needle.size: raise ValueError("[ERROR in gu.mask] cross_impose must have equal size to needle or be a singlular item")
        
        mask_invert = np.ones(haystack.size, dtype=np.bool)
        mask_invert[idx] = 0
        idx_anti = np.arange(haystack.size)[mask_invert]

        haystack_copy = haystack.copy()
        haystack_copy.flat[idx] = impose
        haystack_copy.flat[idx_anti] = cross_impose
        
        return haystack_copy
    else:    
        return idx

def comask(haystack1, haystack2):
    """similar to mask but makes sure exclusivity in both haystacks. mask might not align
    both arrays together"""
    
    idx = np.searchsorted(haystack1, haystack2)
    mask = idx < haystack1.size
    mask[mask] = haystack1[idx[mask]] == haystack2[mask]
    idx1 = idx[mask]
    
    idx = np.searchsorted(haystack2, haystack1)
    mask = idx < haystack2.size
    mask[mask] = haystack2[idx[mask]] == haystack1[mask]
    idx2 = idx[mask]
    
    return idx1, idx2
    
def overlaps(arr, invert=False):
    """Checks for overlapping rows"""
    
    #Sort array by small difference between elements
    arr = arr[np.argsort(np.diff(arr,axis=1).ravel())]

    _todelete = []
    for i in xrange(arr.shape[0]): 
        _todelete.append(bool2int((arr[i,0] >= arr[:,0]) & (arr[i,1] <= arr[:,1]))[1:])
    
    todelete = unique(flatten(_todelete)).astype(int)
    if invert is not False: todelete = np.arange(arr.shape[0])[~np.in1d(np.arange(arr.shape[0]), todelete)]
    
    arr = np.delete(arr, todelete, axis=0)

    return arr
        
def flat_old(array):
    """
    Inner function for flatten. Returns original array if array is already
    flat.
    """
    
    try:
        if isinstance(array, list):
            return [item for sublist in array for item in sublist]
        elif isinstance(array, np.ndarray):
            return np.array([item for sublist in array for item in sublist])
        elif isinstance(array, tuple):
            return tuple([item for sublist in array for item in sublist])
        else:
            raise ValueError("flat requires either list, tuple or ndarray")
            
    except TypeError: # Assumption: when array is already flat
        return array

def flat(array, level=1):
    """
    Flatten a list of (lists of (lists of strings)) for any level 
    of nesting
    
    Parameters
    ----------
    array : array_like, string
        An array_like object containing all string values.
    level : int, optional
        The number of nested levels to remove. Specify -1 to removed all
        nests automatically.
        
    Returns
    -------
    x : generator
        The flattened array is returned as a generator. Use as an 
        iterator, or used list(x) to return values.
    
    Reference
    ---------
    https://stackoverflow.com/a/17864492
    https://stackoverflow.com/a/5286571/8765762
    
    Notes
    -----
    Only works with both Python 2.x and 3.x
    
    """
    
    for x in array:
        
        # If desired level of nesting reached, yield x
        if level == 0:
            yield x
        
        # Else if not a string level, go down another level
        elif hasattr(x, '__iter__') and not isinstance(x, str):
            for y in flat(x, level=level-1):
                yield y

        # If at string level, then yield x
        else:
            yield x

def flatten(array, type=None, dtype=None, level=1, generator=False):
    """
    Flattens the input array into 1D.
    
    type : 'list' or 'ndarray'
    dtype : when ndarray is used, specify the dtype or it will be found automatically
    level : number of levels to flatten. Option is -1
    """
    
    # NEW: Flatten the array recursively using flat function.
    array = flat(array, level=level)
    
    if generator is False:
        array = list(array)
        
        if type is not None:
            if type == 'list':
                return array
            elif type == 'ndarray':
                if dtype is None:
                    if isinstance(array, np.ndarray):
                        try:
                            return np.array(array, dtype=array.dtype)
                        except ValueError:
                            UserWarning('[Error] Cannot convert to NumPy. Returning as a list instead')
                            return array
                    else:
                        warnings.warn('\n[flatten]: A dtype must be specified if the input array is a list and you want the output array to be numpy. Numpy will estimate the data type automatically!', SyntaxWarning, stacklevel=2)
                        return np.array(array)
                else:
                    try:
                        return np.array(array, dtype=dtype)
                    except ValueError:
                        UserWarning('[Error] Cannot convert to NumPy. Returning as a list instead')
                        return array
            else:
                UserWarning('[Error] Input array to flatten must be either a list or ndarray. Returning empty array.')
                return []
        else:
            if isinstance(array, list):
                return array
            elif isinstance(array, np.ndarray):
                if dtype is None:
                    try:
                        return np.array(array, dtype=array.dtype)
                    except ValueError:
                        UserWarning('[Error] Cannot convert to NumPy. Returning as a list instead')
                        return array
                else:
                    try:
                        return np.array(array, dtype=dtype)
                    except ValueError:
                        UserWarning('[Error] Cannot convert to NumPy. Returning as a list instead')
                        return array
            else:
                UserWarning('[Error] Input array to flatten must be either a list or ndarray. Returning empty array.')
                return []
    else:
        return array

def flatten_old(array, type=None, dtype=None, level=1):

    level -= 1

    array_old = deepcopy(array)
    array = flat(array)
    
    # If level != 0, perform recursion!
    if level == 0:
        return_data = True
        
    # If level == -1, perform auto flattening to bare bones
    elif (level < 0) and np.array_equal(array, array_old):
        return_data = True
    
    # Escape when recursion is excessive.
    elif level < -50:
        return_data = True
    
    # Recursion!
    else:
        return flatten(array, type=type, dtype=dtype, level=level)
    
    if return_data:
        if type is not None:
            if type == 'list':
                return array
            elif type == 'ndarray':
                if dtype is None:
                    if isinstance(array, np.ndarray):
                        try:
                            return np.array(array, dtype=array.dtype)
                        except ValueError:
                            UserWarning('[Error] Cannot convert to NumPy. Returning as a list instead')
                            return array
                    else:
                        warnings.warn('\n[flatten]: A dtype must be specified if the input array is a list and you want the output array to be numpy. Numpy will estimate the data type automatically!', SyntaxWarning, stacklevel=2)
                        return np.array(array)
                else:
                    try:
                        return np.array(array, dtype=dtype)
                    except ValueError:
                        UserWarning('[Error] Cannot convert to NumPy. Returning as a list instead')
                        return array
            else:
                UserWarning('[Error] Input array to flatten must be either a list or ndarray. Returning empty array.')
                return []
        else:
            if isinstance(array, list):
                return array
            elif isinstance(array, np.ndarray):
                if dtype is None:
                    try:
                        return np.array(array, dtype=array.dtype)
                    except ValueError:
                        UserWarning('[Error] Cannot convert to NumPy. Returning as a list instead')
                        return array
                else:
                    try:
                        return np.array(array, dtype=dtype)
                    except ValueError:
                        UserWarning('[Error] Cannot convert to NumPy. Returning as a list instead')
                        return array
            else:
                UserWarning('[Error] Input array to flatten must be either a list or ndarray. Returning empty array.')
                return []
                
def near(array,value):
    """https://stackoverflow.com/a/2566508"""

    idx = np.abs(array-value).argmin()
    
    return array[idx]

def argnear(array, value):
    """based on find_nearest but returns index rather than the array value"""
    
    return int(np.abs(array-value).argmin())

def argneararray(array, values):
    """Same as argnear but can accepts both arrays and single values"""
    
    if isinstance(values, list) or isinstance(values, np.ndarray):
        return np.argmin([np.square(array-value) for value in values], axis=1)
    else:
        return argnear(array, values)
        
def round(a, clip):
    """Will round number, a, to closes multiple of clip.
    e.g. round(5.9, 12) = 0
         round(6.0, 12) = 12.0
         
    Ref: https://stackoverflow.com/a/7859208/8765762
    """
    
    return np.round(float(a) / clip) * clip    
    
def ceil(a, clip=1):
    """Same as round but will raise the value"""

    return np.ceil(float(a) / clip) * clip    
    
def floor(a, clip=1):
    """Same as round but will floor the value"""

    return np.floor(float(a) / clip) * clip    

def truncate(a, floor=True):
    """
    truncates a to the nearest order of magnitude.
    
    Parameters
    ----------
    a : numerical
        A single numerical value to truncate
    floor : boolean, optional, default is True
        Specifiy whether to floor the result (True) or to ceiling the 
        result (False). See examples for use.
    
    Examples
    --------
    >>> a = 0.0004334454
    >>> gu.truncate(a)
    0.0001
    
    >>> a = 9567.34
    >>> gu.truncate(a)
    1000
    >>> gu.truncate(a, floor=False)
    10000
    
    """
    
    # Error check
    if not gilly_utilities.system.isnumeric(a): 
        raise ValueError("truncate requires a single numerical value")
    
    # Truncate the input numeric
    if floor:
        return 10**int(('%e' % a)[-3:])
    elif not floor:
        return 10**int(('%e' % a)[-3:]) * 10
    else:
        raise ValueError("floor must be a boolean. See docstring for info.")

def bool2int(bool, keepfalse=False, falseval=0):
    """Converts a boolean array into the integer counterparts. So for every True value, the
    element will be returned as an array ready for indexing. 
    
    Parameters
    ----------
    
    bool : ndarray, boolean
        an ndarray of true and false conditions relating to some argument."""
    
    bool = np.asarray(bool)
    
    if bool.dtype == 'bool':
        if keepfalse is False:
            return np.arange(bool.size, dtype=int)[bool]
        else:
            arr = np.arange(bool.size, dtype=int)
            try:
                arr[~bool] = falseval
            except:
                arr = arr.astype(float)
                arr[~bool] = falseval
            
            return arr
    else:
        raise ValueError("[bool2int] input array did not have a dtype bool. We got %s" % bool.dtype)

def int2bool(arr, int):
    """Converts an integer array (typically used for indexing) to be used as a boolean array. This can
    be useful when subset an ND array.
    
    Parameters
    ----------
    arr : 1D array
        a 1 dimensional array used for dimension keeping
    int : 1D array
        a 1 dimensional integer array used to select the elements of arr.
        
    N.B. Output will be a boolean array of size arr. This can be used to select the elements in arr in the
    same way as just using the int array."""
    
    # Create False array with size equalling arr
    arr_bool = np.full(arr.size, False)
    
    # For all index values, set arr_bool to True
    arr_bool[int] = True
    
    return arr_bool

def combine_bools(arr):
    """
    Combines all boolean arrays in arr which all have the same length.
    If any array has False, the output will also be False.
    
    >>> a = [True, False, True, True]
    >>> b = [False, True, True, True]
    >>> combine_bools((a, b))
    array([False, False, True, True])
    
    """
    
    bools = ~np.asarray(arr[0])
    for elem in arr[1:]:
        bools |= ~np.asarray(elem)
     
    return ~bools
    
def strip(array, mask=""):
    """Removes all mask values in an array.
    
    Parameters
    ----------
    array : list or ndarray
        The list or ndarray that you want to remove values from
    mask : value, list or ndarray
        Input the values you want to mask e.g. mask=1, mask=[1,5]
        mask = np.array([1,5])
        
    Return
    ------
    array_masked : ndarray
        An ndarray which have the masked values removed.
        
    Notes
    -----
    No matter what type of input array you use (e.g. list) the output
    array will always be a numpy ndarray. numpy arrays are easier to
    remove multiple values from arrays.
    
    Examples
    --------
    
    len(Data['FieldMill_Chilbolton_1sec_Step_File'][Data['FieldMill_Chilbolton_1sec_Step_File'] != ""])
        
    """
    
    array = np.array(array)
    
    if isinstance(mask, np.ndarray) or isinstance(mask, list):
        ind_mask = np.array([array.T != val for val in mask]).T
        all_mask = np.array([all(tup) for tup in ind_mask], dtype=bool)
        #all_mask = [all(tup) for tup in ind_mask]
        
        try:
            return array[all_mask]
        except IndexError:
            return np.array([])
    else:
        return array[array != mask]

def contiguous(x, min=1, invalid=None, bounds=False):
    """Locate regions of contiguous data and return an array of the same shape as input with ID.
    
    e.g. [0,0,0,0,1,1,1,1,0,0,0,0,2,2,2,2] 
    
    where 0 is not a valid data location and 1,2 are unique areas
    
    Parameters
    ----------
    x : ndarray,
        A 1D numpy array of values that have already been pre-formatted. For example not valid elements
        must have a np.nan or np.inf to recognise this as an out-of-bounds areas
    min : integer, optional, default=1
        An integer that specifies the minimum length that a contiguous area can be in length
    invalid : any, optional, default=None
        A value specifying the invalid value. E.g. invalid=0 will see the value 0 as the spacing between
        any possible contiguous region. Default is None which will use np.finite to find any invalid values.
    bounds : bool, optional. default=False
        Specify if you want to return the index boundaries of each contiguous region
    """
    
    x_mask = np.zeros(x.size, dtype=int)
    Temp_Mask = 0
    Ind_Mask = ()
    if invalid is None:
        for j in xrange(x.size):
            if np.isfinite(x[j]):
                Temp_Mask += 1
                Ind_Mask += (j,)
                x_mask[(Ind_Mask,)] = Temp_Mask
            else:
                Temp_Mask = 0
                Ind_Mask = ()
    else:
        for j in xrange(x.size):
            if x[j] != invalid:
                Temp_Mask += 1
                Ind_Mask += (j,)
                x_mask[(Ind_Mask,)] = Temp_Mask
            else:
                Temp_Mask = 0
                Ind_Mask = ()        
    
    #Remove contiguous lengths less than min
    x_mask[x_mask < min] = 0
            
    #Convert integers from cloud length to unique cloud number
    x_bounds = bool2int(np.diff(x_mask) != 0) + 1
    
    #Check to see if cloud exists along day boundary. If so we add in manual time boundaries
    if x_mask[0] != 0: x_bounds = np.insert(x_bounds, 0, 0)
    #if x_mask[-1] != 0: x_bounds = np.append(x_bounds, x_mask.size); 
    if x_mask[-1] != 0: x_bounds = np.append(x_bounds, x_mask.size-1); 
    #if x_mask[-1] != 0: x_bounds = np.append(x_bounds, Temp_Mask); 
    x_bounds = x_bounds.reshape(int(x_bounds.size/2),2)

    #Give each cloud a unique integer for each time step 
    for cloud_id, cloud_bounds in enumerate(x_bounds,1):
        x_mask[cloud_bounds[0]:cloud_bounds[1]] = cloud_id
    
    if bounds is True:
        return x_bounds, x_mask
    elif bounds is False:
        return x_mask

def nearcontiguous(x, min=1, min_spacing=2, invalid=None, bounds=False):
    """Find regions of data which are near contiguous. You can specify the maximum spacing between
    contiguous regions and the minimum spacing between non-contiguous regions. The maximum spacing
    between contiguous regions must be smaller than the minimum spacing between non-contiguous 
    regions.
    
    Parameters
    ----------
    x : ndarray,
        A 1D numpy array of values that have already been pre-formatted. For example not valid elements
        must have a np.nan or np.inf to recognise this as an out-of-bounds areas
    min : integer, optional, default=1
        An integer that specifies the minimum length that a contiguous area can be in length
    min_spacing : integer, optional, default=2
        An integer that specifies the minimum spacing between contiguous areas
    invalid : any, optional, default=None
        A value specifying the invalid value. E.g. invalid=0 will see the value 0 as the spacing between
        any possible contiguous region. Default is None which will use np.finite to find any invalid values.
    bounds : bool, optional. default=False
        Specify if you want to return the index boundaries of each contiguous region
    
    Example
    -------
    
    >>> x = np.array([0,0,0,0,0,0,1,0,1,0,1,1,1,0,1,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,0,1,0,0,1,0,0,0,1,
                1,0,0,0,0,0,0,0,0,1,0,1,0,0,0,1,0,0,0,0,1,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,1,0,0,1,1,
                1,0,1,0,0,0,0,0,0,0])
    >>> gu.nearcontiguous(x, min=2, min_spacing=6, invalid=0)
    [[6,14],[35,43],[52,63],[84,91]]
    
    """
    
    #Force x to be a numpy array. Create a sister variable to capture index values.
    x = np.asarray(x).copy()
    sister = np.arange(x.size)
    
    check_contiguous = True
    while check_contiguous is True:
        #Calculate the time between tips
        time_spacing = np.diff(sister[x != invalid])
        
        #Find locations were time between tip is greater than threshold
        event_mask = time_spacing >= min_spacing
        
        #Look for contiguous areas. (i.e. if a rain event only has one tip)
        contiguous_mask = bool2int(contiguous(event_mask, min=min, invalid=invalid) != 0)[1::2]
        
        #If contiguous data found, remove from original rain data
        if contiguous_mask.size != 0:
            x[sister[x != invalid][contiguous_mask]] = 0
            check_contiguous = True
        else:
            check_contiguous = False
            event_mask = bool2int(event_mask)

    #Add value to start and end of x to capture the first and last near-contiguous events
    x_large = np.append(np.insert(x,0,np.append(1,np.zeros(min_spacing))), np.append(np.zeros(min_spacing),1))
    sister_large = np.arange(x_large.size)
    
    #Create reversed arrays to capture start and end of boundaries
    x_reverse = x_large[::-1]
    sister_reverse = sister_large[::-1]
    
    #Find the index value for the end boundary point
    endpoint = sister_large[x_large != invalid][bool2int(np.diff(sister_large[x_large != invalid]) >= min_spacing)]
    
    #Find the index value for the start boundary point
    startpoint = sister_reverse[x_reverse != invalid][bool2int(np.diff(sister_reverse[x_reverse != invalid]) <= -min_spacing)]
    
    #Join index together
    index = np.sort(np.append(startpoint, endpoint), kind='mergesort')[1:-1]
    
    #Reshape index and remove 1 from each index to compensate for the extra value added earlier
    index = index.reshape(int(index.size/2), 2) - min_spacing - 1
    
    #Create array with same size as x with the location of the near-contiguous regions
    val = 1
    x_output = np.zeros(x.size, dtype=int)
    for region in index:
        x_output[region[0]:region[1]+1] = val
        val += 1
    
    return index, x_output

def argcontiguous(x, min=1, valid=None, bounds=False):
    """Creates the anti array of contiguous"""
    
    #First get the mask from contiguous
    mask = contiguous(x, min=min, invalid=valid, bounds=False)
    
    #Next, convert mask to floats (incase valid is nan or inf)
    mask = mask.astype(np.float64)
    
    #Check if valid parameter is the same as non values in mask
    if valid == 0: 
        mask[mask == 0] = -1
        check = -1
    else:
        check = 0
    
    #Make all true values in mask valid
    mask[mask != check] = valid
    mask[mask == check] = 1
    
    #Return the anticontiguous
    return contiguous(mask, min=min, invalid=valid, bounds=bounds)
    
def argnearcontiguous(x, min=1, max_spacing=2, valid=None, bounds=False):
    """Creates the anti array of contiguous"""
    
    #First get the mask from contiguous
    mask = nearcontiguous(x, min=min, min_spacing=max_spacing, invalid=valid, bounds=False)[1]
    
    #Next, convert mask to floats (incase valid is nan or inf)
    mask = mask.astype(np.float64)
    
    #Check if valid parameter is the same as non values in mask
    if valid == 0: 
        mask[mask == 0] = -1
        check = -1
    else:
        check = 0
    
    #Make all true values in mask valid
    mask[mask != check] = valid
    mask[mask == check] = 1
    
    #Return the anticontiguous
    return nearcontiguous(mask, min=min, min_spacing=max_spacing, invalid=valid, bounds=bounds)
    
def indexswap(arr, ind, swaparr):
    """Swaps index array that are linked with arr, to a different array, swaparr"""
    
    if len(ind) == 0: return np.array([])
    
    #Get values from array
    val = arr[ind]
        
    #Find closest value in swaparr
    swapind = argneararray(swaparr, val.ravel()).reshape(val.shape)
    
    return swapind
        
def searchsorted(a, v, side='left', sorter=None):
    """Identical to the numpy version of search sorted except we take care of the
    out-of-bounds indices if a value in 'v' is outside all values in 'a'"""
    
    a = np.asarray(a)
    
    #Get indices from numpy search sorted
    ind = np.searchsorted(a, v, side=side, sorter=sorter)
    
    #Check that ind is an array. If only one element was found, np.searchsorted returns an scalar rather than an array.
    if not isinstance(ind, np.ndarray): ind = np.asarray(ind, dtype=int)
    
    #For any indices equalling the size of the input array, a, we remove 1 from the index
    if ind.size != 0: ind[ind == a.size] -= 1
    
    return ind    
    
def digitize(a, v, right=False):
    """Identical to the numpy version of digitize except we take care of the
    out-of-bounds indices if a value in 'v' is outside all values in 'a'"""
    
    a = np.asarray(a)
    
    #Get indices from numpy search sorted
    ind = np.digitize(a, v, right=right)
    
    #Check that ind is an array. If only one element was found, np.searchsorted returns an scalar rather than an array.
    if not isinstance(ind, np.ndarray): ind = np.asarray(ind, dtype=int)
    
    #For any indices equalling the size of the input array, a, we remove 1 from the index
    if ind.size != 0: ind[ind == v.size] -= 1
    
    return ind    
        
def broadcast(array, window, step=1, undersample=False):  # Window len = L, Stride len/stepsize = S
    """Restructures array into rows of length, L with step sizes, S.
    
    Parameters
    ----------
    array : 1D array
        The array you want to restructure
    window : int
        The number of elements per row
    step : int, optional
        The number of steps you want in the column
    under-sample : boolean, optional
        Specify if you want to broadcast the array to show the under-sample bins.
    Example
    -------
    
    >>> array = np.arange(10)
    >>> gu.broadcast(array, 5, 2)
    array([[0, 1, 2, 3, 4],
       [2, 3, 4, 5, 6],
       [4, 5, 6, 7, 8]])
    
    Reference
    ---------
    https://stackoverflow.com/questions/40084931/taking-subarrays-from-numpy-array-with-given-stride-stepsize/40085052#40085052
    
    Update
    ------
    Reformatted output array when undersample is True. Now we maintain a perfect NxM array and pad 
    any undersampled array with np.nan. NOTE TO SELF: why did we not do this originally?
    """
    
    #Check is step if an int or float
    if isinstance(step, float): step = int(step)
    if not isinstance(step, int): step = 1
        
    if undersample is False:
        
        nrows = ((array.size-window)//step)+1
        return array[step*np.arange(nrows)[:,None] + np.arange(window)]
    
    elif undersample is True:
        #bins = np.array([array[i if i>0 else 0:i+window].tolist() for i in xrange(-window+1,array.size)])[0::step]
        bins = np.array([array[i if i>0 else 0:i+int(window)].tolist() for i in xrange(-int(window)+1,array.size)])[0::int(step)]
        
        # binslen_before = np.zeros(len(bins), dtype=int)
        # for i, data in enumerate(bins):
            # binslen_before[i] = len(data)
        
        #Pad array with np.nans for under sampled bins
        for i, data in enumerate(bins):
            bins[i] = np.pad(data, (window-len(data),0), 'constant', constant_values=np.nan)

        #flatten array to remove object identifiers
        bins = flatten(bins)
        
        #Convert to float and reshape to match desired arrangement
        bins = np.array(bins, dtype=float).reshape(int(len(bins)/window), window)
        
        # binslen_after = np.zeros(len(bins), dtype=int)
        # for i, data in enumerate(bins):
            # binslen_after[i] = len(data)
        
        # print("binslen_before", binslen_before.tolist())
        # print("binslen_after", binslen_after.tolist())
        # print("Length", len(binslen_before), binslen_after.shape)
        # sys.exit()
        return bins

def stride(array1, window, step=1, undersample=False): # Window len = L, Stride len/stepsize = S
    """Restructures array into rows of length, L with step sizes, S. This is similar to 
    the broadcast function but numpy as_strided is much faster as it reorganise the 
    memory of the input array rather than forming a new array.
    
    DANGER
    ------
    THIS FUNCTION CAN BE VOLITALE. OVERUSE OF THIS FUNCTION CAN LEAK MEMORY AND CAUSE MEMORY
    OVERFLOW ISSUES. USE AT YOUR OWN PERIL.
    
    Parameters
    ----------
    array : 1D array
        The array you want to restructure
    window : int
        The number of elements per row
    step : int, optional
        The number of steps you want in the column
    under-sample : boolean, optional
        Specify if you want to broadcast the array to show the under-sample bins.
    Example
    -------
    
    >>> array = np.arange(10)
    >>> gu.broadcast(array, 5, 2)
    array([[0, 1, 2, 3, 4],
       [2, 3, 4, 5, 6],
       [4, 5, 6, 7, 8]])
    
    Reference
    ---------
    https://stackoverflow.com/questions/40084931/taking-subarrays-from-numpy-array-with-given-stride-stepsize/40085052#40085052
    
    """
    
    #Check is step if an int or float
    if isinstance(step, float): step = int(step)
    if not isinstance(step, int): step = 1
    
    nrows = ((array1.size-window)//step)+1
    n = array1.strides[0]
    
    if undersample is False:
        
        return np.lib.stride_tricks.as_strided(array1, shape=(nrows,window), strides=(step*n,n))
        
    elif undersample is True:
        
        t1 = systime.time()
        
        mid = np.lib.stride_tricks.as_strided(array1, shape=(nrows,window), strides=(step*n,n))
        
        t2 = systime.time()
        
        data = array()
        for i in xrange(1,window):
            temp = mid[0].copy()
            temp[i:] = np.nan
            data.update([temp])
        
        t3 = systime.time()

        for elements in mid: data.update([elements])
        
        t4 = systime.time()
        
        for i in xrange(1,window):
            temp = mid[-1].copy()
            temp[:i] = np.nan
            data.update([temp])

        t5 = systime.time()
        
        data = data.finalize(flat=False)
        
        t6 = systime.time()
        #print("t2-t1 = %.6fs, t3-t2 = %.6fs, t4-t3 = %.6fs, t5-t4 = %.6fs, t6-t5 = %.6fs, total = %.6fs" % (t2-t1, t3-t2, t4-t3, t5-t4, t6-t5, t6-t1))
        
        return data
    
def unique(ar, return_index=False, return_inverse=False,
            return_counts=False, axis=None):
    """
    Find the unique elements of an array.
    Returns the sorted unique elements of an array. There are three optional
    outputs in addition to the unique elements: the indices of the input array
    that give the unique values, the indices of the unique array that
    reconstruct the input array, and the number of times each unique value
    comes up in the input array.
    
    Part of numpy version 1.14.0. The code was ripped from
    https://github.com/numpy/numpy/blob/v1.14.0/numpy/lib/arraysetops.py#L113-L248
    """
    
    if np.asarray(ar).size == 0: return np.asarray(ar)
    
    ar = np.asanyarray(ar)
    if axis is None:
        return _unique1d(ar, return_index, return_inverse, return_counts)
    if not (-ar.ndim <= axis < ar.ndim):
        raise ValueError('Invalid axis kwarg specified for unique')

    ar = np.swapaxes(ar, axis, 0)
    orig_shape, orig_dtype = ar.shape, ar.dtype
    # Must reshape to a contiguous 2D array for this to work...
    ar = ar.reshape(orig_shape[0], -1)
    ar = np.ascontiguousarray(ar)

    if ar.dtype.char in (np.typecodes['AllInteger'] +
                            np.typecodes['Datetime'] + 'S'):
        # Optimization: Creating a view of your data with a np.void data type of
        # size the number of bytes in a full row. Handles any type where items
        # have a unique binary representation, i.e. 0 is only 0, not +0 and -0.
        dtype = np.dtype((np.void, ar.dtype.itemsize * ar.shape[1]))
    else:
        dtype = [('f{i}'.format(i=i), ar.dtype) for i in range(ar.shape[1])]

    try:
        consolidated = ar.view(dtype)
    except TypeError:
        # There's no good way to do this for object arrays, etc...
        msg = 'The axis argument to unique is not supported for dtype {dt}'
        raise TypeError(msg.format(dt=ar.dtype))

    def reshape_uniq(uniq):
        uniq = uniq.view(orig_dtype)
        uniq = uniq.reshape(-1, *orig_shape[1:])
        uniq = np.swapaxes(uniq, 0, axis)
        return uniq

    output = _unique1d(consolidated, return_index,
                        return_inverse, return_counts)
    if not (return_index or return_inverse or return_counts):
        return reshape_uniq(output)
    else:
        uniq = reshape_uniq(output[0])
        return (uniq,) + output[1:]

def _unique1d(ar, return_index=False, return_inverse=False,
              return_counts=False):
    """
    Find the unique elements of an array, ignoring shape.
    
    Part of numpy version 1.14.0. The code was ripped from
    https://github.com/numpy/numpy/blob/v1.14.0/numpy/lib/arraysetops.py#L250-L295
    
    This is needed for unique function to work
    """
    ar = np.asanyarray(ar).flatten()

    optional_indices = return_index or return_inverse
    optional_returns = optional_indices or return_counts

    if ar.size == 0:
        if not optional_returns:
            ret = ar
        else:
            ret = (ar,)
            if return_index:
                ret += (np.empty(0, np.intp),)
            if return_inverse:
                ret += (np.empty(0, np.intp),)
            if return_counts:
                ret += (np.empty(0, np.intp),)
        return ret

    if optional_indices:
        perm = ar.argsort(kind='mergesort' if return_index else 'quicksort')
        aux = ar[perm]
    else:
        ar.sort()
        aux = ar
    flag = np.concatenate(([True], aux[1:] != aux[:-1]))

    if not optional_returns:
        ret = aux[flag]
    else:
        ret = (aux[flag],)
        if return_index:
            ret += (perm[flag],)
        if return_inverse:
            iflag = np.cumsum(flag) - 1
            inv_idx = np.empty(ar.shape, dtype=np.intp)
            inv_idx[perm] = iflag
            ret += (inv_idx,)
        if return_counts:
            idx = np.concatenate(np.nonzero(flag) + ([ar.size],))
            ret += (np.diff(idx),)
    return ret        

def unpack(*arg):
    """Unpacks a 2D array. Input must be unpack(*arg.T) for column unpacking and
    unpack(*arg) for row unpacking."""
    
    return arg

def fix_unicode(data, encoding='utf-8'):
    """Fixes strings with unicode parts (e.g. u'hello' --> 'hello')
    
    Reference
    ---------
    https://stackoverflow.com/a/18273297/8765762
    """
    
    if isinstance(data, unicode):
        return data.encode(encoding)
    elif isinstance(data, dict):
        data = dict((fix_unicode(k), fix_unicode(data[k])) for k in data)
    elif isinstance(data, list):
        for i in xrange(0, len(data)):
            data[i] = fix_unicode(data[i])
    return data
    
def fix_dtypes(dtype):
    """Fixes an issue with pandas when using pd.DataFrame.to_records which will
    return a np.recarray with unicode dtypes. The unicode dtypes are
    incompatible with numpy so an manipulation of the record array thereforth
    is extremely difficult
    
    Reference
    ---------
    https://stackoverflow.com/a/18273297/8765762
    https://stackoverflow.com/a/11108384/8765762
    
    """
    
    #First, need to convert [(,),(,)] into [[],[]] structure
    dtype = [list(line) for line in dtype]
    
    #Second, using fix_unicode, remove all unicode parts in string (e.g. u'hello' --> 'hello')
    return [tuple(line) for line in fix_unicode(dtype)]
    
def fix_recarray(recarray):
    """Building upon fix_dtypes, this function will convert a corrupted np.recarray with incorrect
    dtypes (e.g. [(u'f0', '<f8'), (u'f0', '<f8')] to [('f0', '<f8'), ('f0', '<f8')]"""
    
    return recarray.astype(fix_dtypes(recarray.dtype.descr))
    
def rec_append(recarray, data, dtype):
    """Adds colomn to np.recarry. This function allows multiple coloumns to be
    addition on one line.
    
    Examples
    --------
    >>> #Add one coloumn to record array with float64 ('<f8') dtype
    >>> self.data = gu.rec_append(self.data, data=PLL, dtype=[('PLL', '<f8')])
    
    >>> #Add multiple coloumns to record array-value
    >>> self.data = gu.rec_append(self.data, data=[dfdt, np.abs(SLWC), z_run, f_run], dtype=[('dfdt', '<f8'),('SLWC','<f8'),('z_run','<f8'),('f_run','<f8')])
    """
    
    #Create new record array with new dtypes appended to end.
    new_dt = np.dtype(fix_dtypes(recarray.dtype.descr) + dtype)
    temp_array = np.zeros(recarray.shape, dtype=new_dt)
    
    #Add previous data to new record array
    for name in recarray.dtype.names: temp_array[name] = recarray[name]
    
    #Add new data to record array
    if len(dtype) > 1:
        for name, dat in zip(np.array(dtype)[:,0], data): 
            temp_array[name] = dat
    else:
        temp_array[dtype[0][0]] = data
        
    return temp_array
    
def string_checker(strings, test_strings, condition='all'):
    """Tests whether any strings are inside other strings!
    
    References
    ----------
    https://stackoverflow.com/a/3389611/8765762
    """
    
    if isinstance(strings, str):
        strings = [strings]
    elif gilly_utilities.system.isarray(strings):
        if len(strings) == 0:
            return False
    else:
        raise ValueError("[gu.string_checker] Incorrect data type")
    try:
        return getattr(np, condition)([[x in strs for x in test_strings] for strs in strings], axis=1)
    except ValueError:
        warnings.warn("[gu.string_checker] We got a ValueError. Most likely because strings has length 0. We'll let you off this once")
        
def dict2array(dic):
    """
    Creates a numpy record array from a python dictionary no matter
    how nested the dictionary is.
    
    N.B. Dictionary needs to be structured with same length for each
    keys
    
    N.B. Each key must have a numpy array embedded inside.
    
    References
    ----------
    https://stackoverflow.com/a/32336904
    """

    def _mkdt(dic):
        """
        Detemines dtypes from dictionary with numpy arrays
        """

        ll = []
        if isinstance(dic, dict):
            for k, v in dic.items():
                if isinstance(v, np.ndarray):
                    ll.append((k, v.dtype))
                else:
                    ll.append((k, _mkdt(v)))
        return ll

    def _mkshape(dic):
        """
        Detemines shape from dictionary with numpy arrays
        """

        ll = []
        if isinstance(dic, dict):
            for k, v in dic.items():
                print(k, v, end="\n\n")
                if isinstance(v, np.ndarray):
                    ll.append((v.shape[0]))
                else:
                    ll.append((_mkshape(v)))
        print("ll", ll)
        return np.hstack(np.array(ll).flat)

    def _copy_values(dic, arr):
        """
        Copy nested values from dic to arr
        """

        if arr.dtype.names:
            for n in arr.dtype.names:
                copy_values(dic[n], arr[n])
        else:
            arr[:] = dic

    #Get shape of dictionary
    shape = _mkshape(dic)
    print("shape", shape)
    #Create blank record array with same shape as dic
    arr = np.zeros(shape, _mkdt(dic))

    #Fill in array with values from dict
    _copy_values(dic, arr)

    return arr

def array2recarray(data, names, formats):
    """
    Creates a record array from a series of numpy arrays
    
    Example
    -------
    >>> a = np.linspace(-100,20,200)
    >>> b = np.full(200,100)
    >>> gu.array2recarray((a,b), names='RH, Tdry', formats='f8, f8')
    
    Reference
    ---------
    https://stackoverflow.com/a/5204280/8765762
    """
    
    return np.core.records.fromarrays(data, names=names, formats=formats)
    
def padbool(mask, padwidth):
    """
    Pads the True elements in mask with a length defined by padwidth.
    
    Example
    -------
    >>> mask = np.array([False, False, True, False, False], dtype=bool)
    >>> gu.padbool(mask, (1,1))
    [False True True True False]
    
    Parameters
    ----------
    mask : array_like
        An array_like object which contains only True or False values
    padwidth : array_like or int
        An array_like specifying the number of elements to pad around
        any True elements with has the format (<left_pad>, <right_pad>)
        
        e.g. padwidth = (1,1) will pad 1 element around each True value
        found.
        
    """
    
    # Ensure mask is ndarray
    mask = np.array(mask, dtype=bool)
    
    # Ensure padwidth is in a consistent format
    if isinstance(padwidth, int) or isinstance(padwidth, float):
        padwidth = (int(padwidth), int(padwidth))
    
    # Convert 
    for elem in bool2int(mask == True):
        mask[elem - padwidth[0]:elem + padwidth[1] + 1] = True
        
    return mask
    
def flip(array):
    """
    Reverses direction of array
    """
        
    return array[::-1]


from __future__ import absolute_import, division, print_function
import sys
import time
import warnings
import matplotlib.pyplot as plt
from matplotlib import transforms
from itertools import takewhile

# Compatability for python3
if getattr(sys.version_info, 'major') == 3:
    xrange = range

#Define alphabet
alphabet = list('abcdefghijklmnopqrstuvwxyz')

# Specify the time when module was initalised. Used in progress if
# t_initial was not specified.
t_initial_master = time.time()

# Used for progress class
eps = sys.float_info.epsilon

class ProgressWarning(UserWarning):
    """Issued when results may be unstable."""
    pass

# On import, make sure that InstabilityWarnings are not filtered out.
warnings.simplefilter('always', ProgressWarning)

class prettyfloat(object):
    """returns a truncated float of 2 decimal places.
    
    Example
    -------
    use: map(prettyfloat, x) for a list, x
    
    Reference
    ---------
    https://stackoverflow.com/a/1567630
    """
    
    def __init__(self, float, dec=6):
        """dec = precision of float"""
        
        self.float=float
        self.dec=dec

    def __repr__(self):
        return "%0.*f" % (self.dec, self.float)

class progress(object):
    """
    Outputs the progress of interable (e.g. loop) to the console.
    
    Examples
    --------
    >>> size = 10
    >>> message = "Hello %s. My name is %s"
    >>> with gu.progress(message, size) as progress:
    >>>     for i in xrange(size):
    >>>         time.sleep(.5)
    >>>         progress.update("James", "Brian")
    """

    def __init__(self, msg, size, confirmation=True):
        """
        Makes a note of the time progress was called.
        
        Parameters
        ----------
        size : int
            the size of the iterator to progress
        """

        # Initalise parameters
        self.t_initial = time.time()
        self.msg = str(msg)
        self.size = int(size)
        self.i = 0
        self.confirmation = confirmation

    def __enter__(self):
        return self
    
    def __exit__(self, type, value, traceback):

        # Remove last line in command line
        sys.stdout.write("\033[K")
        
        # Output the final message to user
        if self.confirmation:
            cprint("[INFO] Iterator Completed (In %.4fs)" \
                    % (time.time() - self.t_initial))
        
    def _timestats(self):
        """
        Calculate the time statistics of the iterator.
        """

        t_elapsed = time.time() - self.t_initial
        t_remain  = (t_elapsed / (self.i + eps)) * \
                    (self.size - self.i + eps)

        return t_elapsed, t_remain

    def _message(self, args):
        """
        Build message to print to output.
        """

        full_message = "[INFO] Time Elapsed: %.0fs. Time Remaining: %.0fs. " \
                         + self.msg + "\r"
        sys.stdout.write(full_message % tuple(args))
        sys.stdout.flush()

    def _error(self):
        """
        Check if iterator has gone above size originally specified.
        """

        # Increment iterator
        self.i += 1

        if self.i > self.size:
            warnings.warn("size of progress was not specified correctly. " \
                          "You'll notice negative values in time remaining!", 
                          ProgressWarning, stacklevel=3)

    def update(self, *args):
        """
        Update the progress of the iterator.
        """

        # Remove last line in command line
        sys.stdout.write("\033[K")

        # Calculate time statistics
        t_elapsed, t_remain = self._timestats()

        # Build message
        self._message(_flat((t_elapsed, t_remain, args)))

        # Error checker
        self._error()

def _flat(array, level=1):
    """
    Flatten a list of (lists of (lists of strings)) for any level 
    of nesting
    
    Parameters
    ----------
    array : array_like, string
        An array_like object containing all string values.
    level : int, optional
        The number of nested levels to remove. Specify -1 to removed all
        nests automatically.
        
    Returns
    -------
    x : generator
        The flattened array is returned as a generator. Use as an 
        iterator, or used list(x) to return values.
    
    Reference
    ---------
    https://stackoverflow.com/a/17864492
    https://stackoverflow.com/a/5286571/8765762
    
    Notes
    -----
    Only works with both Python 2.x and 3.x
    
    """
    
    for x in array:
        
        # If desired level of nesting reached, yield x
        if level == 0:
            yield x
        
        # Else if not a string level, go down another level
        elif hasattr(x, '__iter__') and not isinstance(x, str):
            for y in _flat(x, level=level-1):
                yield y

        # If at string level, then yield x
        else:
            yield x


def inheritors(klass):
    subclasses = set()
    work = [klass]
    while work:
        parent = work.pop()
        for child in parent.__subclasses__():
            if child not in subclasses:
                subclasses.add(child)
                work.append(child)
    return list(subclasses)
   
def rainbow_text(x, y, strings, colors, ax=None, **kw):
    """
    Take a list of ``strings`` and ``colors`` and place them next to each
    other, with text strings[i] being shown in colors[i].

    This example shows how to do both vertical and horizontal text, and will
    pass all keyword arguments to plt.text, so you can set the font size,
    family, etc.

    The text will get added to the ``ax`` axes, if provided, otherwise the
    currently active axes will be used.
    """

    if ax is None:
        ax = plt.gca()
    t = ax.transData
    canvas = ax.figure.canvas

    # horizontal version
    for s, c in zip(strings, colors):
        text = ax.text(x, y, s + " ", color=c, transform=t, **kw)
        text.draw(canvas.get_renderer())
        ex = text.get_window_extent()
        t = transforms.offset_copy(text._transform, x=ex.width, units='dots')

    # vertical version
    for s, c in zip(strings, colors):
        text = ax.text(x, y, s + " ", color=c, transform=t,
                        rotation=90, va='bottom', ha='center', **kw)
        text.draw(canvas.get_renderer())
        ex = text.get_window_extent()
        t = transforms.offset_copy(text._transform, y=ex.height, units='dots')
        
def cprint(*args, **kwargs):
    """Prints colourful messages under a standard style set.
    
    Update: Now handles in the same as does print, meaning you
    can give print as many input strings as you want. e.g.
        
    >>> cprint("Hello", "Brian", "Simon")
    Hello Brian Simon    #but in bold
    
    >>> print("Hello", "Brian", "Simon")
    Hello Brian Simon
    
    No need to specifiy kind anymore. Now defaults to bold.
    Also give kwargs to the print statement now such as 'file'.
    N.B. supplying end in cprint will cause an error as that is
    already being used in this function when print is called."""
    
    #Console Colours
    bcolours = {
        "header" : '\033[95m',
        "okblue" : '\033[94m',
        "okgreen" : '\033[92m',
        "warning" : '\033[31m',
        "fail" : '\033[91m',
        "endc" : '\033[0m',
        "bold" : '\033[1m',
        "underline" : '\033[4m'}
    
    type = kwargs.pop('type', "bold")
    
    for message in args:
        print(bcolours[type] + message + bcolours['endc'], end=' ', **kwargs)
    print("\r")
    
def rgb2hex(r,g,b):
    """Converts RGB values into Hexadecimal
    
    Reference
    ---------
    
    https://stackoverflow.com/a/43572620/8765762
    """
    
    return "#{:02x}{:02x}{:02x}".format(r,g,b)        

def readcomments(filename, comment='#'):
    """
    Reads in filename and extracts only data that contains the comments
    parameter.
    
    Reference
    ---------
    https://stackoverflow.com/a/39724905/8765762
    """
            
    comments_all = []
    with open(filename,'r') as cmt_file:    # open file
        for line in cmt_file:    # read each line
            if line[0] == comment:    # check the first character
                comments_all.append(line[1:])    # remove first '#'

    return comments_all    




from __future__ import absolute_import, division, print_function
import numpy as np
import sys, warnings
from datetime import datetime

from .datetime64 import dt2hours, hours2dt
from .manipulation import bool2int, argnear, searchsorted, contiguous
from .plotting import backend_changer

import matplotlib.pyplot as plt 

# Compatability for python3
if getattr(sys.version_info, 'major') == 3:
	xrange = range
	
def rain_rate(time, counts_or_tips, tip_bucket_size=None, 
        drop_counter_size=None, drop_size=None, mask=True):
    """
    This universal function can convert both tipping bucket, drop 
    counter and disdrometer rain gauge data into a rain rate.
    
    Parameters
    ----------
    time : list or array
        The time values in either hour fraction, datetime or 
        np.datetime64 format. If any other float or integer specification of time is used (e.g. day 
        fraction or second) the outputted rain rate will need to corrected. The
        time is expected to be sorted.
    counts_or_tips : list or array
        The number of counts or tip identifier with the same length as time. For
        tips the input list/array needs to be 0 for no tip and 1 for tip measured.
    tip_bucket_size : float, optional
        The size of the tipping bucket per tip in mm. Default = 0.2 mm
    drop_counter_size : float, optional
        The size of each drop in a drop counter. Default is 0.0033 mm
    drop_size : list or array, optional
        Used to specify the drop sizes of the disdrometer. This must be a 1D array.
    
    Output
    ------
    Time_Tip : numpy array
        The times values when a tip was detected.
    Time_RR : numpy array
        The median time between tips. This represents the central time when the rain
        occurred and naturally has the same number of elements as Rain_RR.
    Rain_RR : numpy array
        The rain rate measured between Time_Tip in mm/hr.
    
    References
    ----------
    Disdrometer : Article
        Islam T. et al. (2012). A Joss-Waldvogel disdrometer derived rainfall 
        estimation study by collocated tipping bucket and rapid response rain gauges. 
        Atmos. Sci. Let. 13, pp 139-150
        
    """
    
    # Check the inputs to function
    if time is None or counts_or_tips is None: 
        return None, None, None
    if not isinstance(time, np.ndarray) or isinstance(time, list): 
        return None, None, None
    if len(time) < 10: 
        return None, None, None
    
    # Force time to be in hours
    time = dt2hours(time, time[0])
    counts_or_tips = np.atleast_1d(counts_or_tips)
    
    if counts_or_tips.ndim == 1:
        if (drop_counter_size is None and drop_size is None) & (np.max(counts_or_tips) == 1 or tip_bucket_size is not None):
            """TIPPING BUCKET"""

            #If the data follows a tipping bucket structure but tip_bucket_size was not specified, then warn user and set to standard 0.2 mm
            if tip_bucket_size is None: 
                warnings.warn('\n[gu.rainrate]: the input data has the same structure as a TIPPING BUCKET but tip_bucket_size was not specified. We will assume a standard bucket size of 0.2 mm!', SyntaxWarning, stacklevel=2)
                tip_bucket_size = 0.2
        
            Rain_Mask = counts_or_tips > 0
            Time_Tip = time[Rain_Mask]
            
            Rain_RR = tip_bucket_size/np.diff(Time_Tip)
            Time_RR = 0.5*np.diff(Time_Tip)+Time_Tip[:-1]
            
            return Time_Tip, Time_RR, Rain_RR
            
        if (tip_bucket_size is None and drop_size is None) & (np.sum(counts_or_tips > 0) > 10):
            """DROP COUNTER"""
            
            if drop_counter_size is None: 
                warnings.warn('\n[gu.rainrate]: the input data has the same structure as a DROP COUNTER but drop_counter_size was not specified. We will assume a standard bucket size of 0.0033 mm!', SyntaxWarning, stacklevel=2)
                drop_counter_size = 0.0033
            
            dt = np.nanmean(np.diff(time))
            
            Rain_Mask = counts_or_tips > 0
            Time_Counts = 0.5*np.diff(time[Rain_Mask])+time[Rain_Mask][:-1] 
            
            
            try:
                Time_Counts = np.append(Time_Counts, Time_Counts[0]-dt)
                Time_Counts = np.append(Time_Counts, Time_Counts[-1]+dt)
            except:
                print("ERROR")
                print("Time_Counts", len(Time_Counts))
                print("time", len(time))
                print("Rain_Mask", Rain_Mask, np.sum(Rain_Mask))
                print("counts_or_tips", counts_or_tips[counts_or_tips != 0])
                print("counts_or_tips", type(counts_or_tips), counts_or_tips.dtype)
                
            Mag_Counts = counts_or_tips[Rain_Mask]
            
            Rain_RR = (drop_counter_size*Mag_Counts)/dt
            Time_RR = time[Rain_Mask]
            
            return Time_Counts, Time_RR, Rain_RR
            
        else:
            #print(44444444444444444)
            return None, None, None
            
    else:
        
        # Transpose counts array
        counts_or_tips = counts_or_tips.T
        
        # Unravel drop size
        drop_size = drop_size.ravel()
        
        if mask is True:
            #try:
            if np.sum(np.sum(counts_or_tips, axis=1) > 1) > 20:
                """DISDROMETER"""
                
                """Calculation method based on Islam T. et al. (2012)"""

                # Constants
                A = 0.005 # m^2
                dt = np.round(np.nanmean(np.diff(time) * 3600)) # ~ 10s
                
                Rain_Mask = np.sum(counts_or_tips, axis=1) > 1
                Time_Disd = 0.5 * np.diff(time[Rain_Mask]) + time[Rain_Mask][:-1] 
                Time_Disd = np.append(Time_Disd, Time_Disd[0] - dt / 3600)
                Time_Disd = np.append(Time_Disd, Time_Disd[-1] + dt / 3600)
                Mag_Disd = counts_or_tips[Rain_Mask]
                        
                #Static Variables (i.e. They DON'T change over each time step)
                v = 3.78*drop_size**0.67
                D = drop_size.copy()**3.67
                dD = 2*np.diff(drop_size); dD = np.append(dD, dD[-1])
                
                #Dynamic Variables (i.e. They DO change over each time step)
                Nm = Mag_Disd/(A*dt*v*dD)
                Rain_RR = (3.78*np.pi*np.nansum(D*Nm*dD, axis=1)/6)/(3600/dt)
                Time_RR = time[Rain_Mask]
                
                #print("Time_Disd", "Time_RR", "Rain_RR")
                #print(Time_Disd.shape, Time_RR.shape, Rain_RR.shape)
                
                return Time_Disd, Time_RR, Rain_RR
                
            else:
                return None, None, None
            # except:
                # print("counts_or_tips", counts_or_tips)
                # print("counts_or_tips", type(counts_or_tips))
                # print("counts_or_tips", len(counts_or_tips))
                # print("counts_or_tips", counts_or_tips.shape)
                # print("counts_or_tips", counts_or_tips.dtype)
        else:
            """DISDROMETER"""
                
            """Calculation method based on Islam T. et al. (2012)"""
            
            #Constants
            A = 0.005 #m^2
            dt = np.round(np.nanmean(np.diff(time)*3600)) #~ 10s
            
            Time_Disd = 0.5*np.diff(time)+time[:-1]; Time_Disd = np.append(Time_Disd, Time_Disd[0]-dt/3600); Time_Disd = np.append(Time_Disd, Time_Disd[-1]+dt/3600)
            Mag_Disd = counts_or_tips
                    
            #Static Variables (i.e. They DON'T change over each time step)
            v = 3.78*drop_size**0.67
            D = drop_size.copy()**3.67
            dD = 2*np.diff(drop_size); dD = np.append(dD, dD[-1])
            
            #Dynamic Variables (i.e. They DO change over each time step)
            Nm = Mag_Disd/(A*dt*v*dD)
            Rain_RR = (3.78*np.pi*np.nansum(D*Nm*dD, axis=1)/6)/(3600/dt)
            Time_RR = time
            
            #print("Time_Disd", "Time_RR", "Rain_RR")
            #print(Time_Disd.shape, Time_RR.shape, Rain_RR.shape)
            
            return Time_Disd, Time_RR, Rain_RR
        
def Drops2RR(drop_time, drop_count, drop_size):
    """Converts a disdrometer drop counter to a rain rate
    
    Paper
    -----
    Islam T. et al. (2012). A Joss-Waldvogel disdrometer derived
    rainfall estimation study by collocated tipping bucket and
    rapid response rain gauges. Atmos. Sci. Let. 13, pp 139-150 
    
    """
    
    #Intialise Equations and Constants
    Nm = lambda n, A, t, v, D: n/(A*t*v*D)
    v = lambda D: 3.78*D**0.67
    RR_Dis_Equ = lambda D, Nm, dD: 3.78*(np.pi/6)*np.sum([D[m]**3.67*Nm[m]*dD[m] for m in xrange(len(D))])
    A = 0.005 #m^2
    dt = 10 #s
    
    #Solve differential drop sizes
    dD = np.zeros(len(drop_size))
    dD[0] = 2*(drop_size[1]-drop_size[0])
    for o in xrange(1, len(drop_size)-1):
        dD[o] = (drop_size[o+1]-drop_size[o])+(drop_size[o]-drop_size[o-1])
    dD[-1] = 2*(drop_size[-1]-drop_size[-2])
    
    #Calculate RR
    RR_Dis = np.zeros(len(drop_time))
    for m in xrange(len(drop_time)):
        RR_Dis[m] = RR_Dis_Equ(drop_size, Nm(drop_count[m], A, dt, v(drop_size), dD), dD)/(3600/dt)
    
    drop_time = drop_time[RR_Dis>0]
    RR_Dis = RR_Dis[RR_Dis>0]
            
    return drop_time, RR_Dis        

def Cloud_Identifer(Time, Sg, Sd):
    """Classifies the type of cloud that is overhead using solar radiation measurements,
    
    Parameters
    ----------
    Time : 1D array of list
        The time for each measurement of Sg and Sd. The time needs to be in fractional hour
    Sg : 1D array or list
        The global solar (shortwave) irradiance
    Sd : 1D array or list
        The global diffuse (longwave) irradiance
        
    Output
    ------
    Cloud_Time : array or float
        A float of the central time for each 15 minute period. If there is more data than
        15 minutes then the output is a float array.
    Cloud_Type : array or int
        An integer specifying the cloud type per 15 minutes of data. If there is more data
        than 15 minutes then the output is an integer array.
    
    """
    
    #Ensure inputs are numpy arrays
    if isinstance(Time[0], np.datetime64):
        Date = Time[0]
        Time = dt2hours(Time, Date)
        
        #print(Date)
        #print(Time)
        #print(Time.size)
        
        Datetime_Found = 'datetime64'
    elif isinstance(Time[0], datetime):
        Date = toDateOnly([Time[-10]])[0]
        Time = np.array([toHourFraction(x) for x in Time], dtype=float)
        Datetime_Found = 'datetime'
    
    else:
        Time = np.array(Time, dtype=float)
        Datetime_Found = 'none'
    Sg = np.array(Sg, dtype=float)
    Sd = np.array(Sd, dtype=float)
    
    #Sort data
    sort_index = np.argsort(Time, kind='mergesort')
    Time = Time[sort_index]
    Sg = Sg[sort_index]
    Sd = Sd[sort_index]
    
    #Determine if Time is shorter than 15 minutes
    if Time[-1] - Time[0] < 0.25: 
        warnings.warn('[Cloud_Identifer]: The identification of cloud type requires 15 minutes or more data. Only %.2f minutes was found!' % ((Time[-1] - Time[0])*60), RuntimeWarning, stacklevel=2)
        return None, None
    
    #Determine if Time is exactly 15 minutes
    elif Time[-1] - Time[0] == 0.25:
        
        Diffuse_Fraction = Sd/Sg

        if np.nanmean(Diffuse_Fraction) >= 0.9:     #Overcast
            Cloud_Type = 2
        elif np.nanmean(Diffuse_Fraction) <= 0.3:     #Clear
            Cloud_Type = 1
        elif np.nanstd(Diffuse_Fraction) <= 0.05:     #Cumuliform
            Cloud_Type = 4
        elif np.nanstd(Diffuse_Fraction) >= 0.1:     #Stratiform
            Cloud_Type = 3
        else:                                    #Unclassified
            Cloud_Type = 5
        
        Cloud_Time = np.nanmean(Time)
    
    #Determine if Time is longer than 15 minutes
    elif Time[-1] - Time[0] > 0.25:
        
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
        
            #Calculate Cloud_Time
            Cloud_Num = int((Time[-1] - Time[0])/0.25)+1
            Cloud_Time = np.array([Time[0] + 0.125 + Time_Step for Time_Step in np.linspace(0, Cloud_Num*0.25, Cloud_Num, endpoint=False)], dtype=float)
            
            #Calculate Cloud_Type for each 15 minute subset
            Cloud_Type = np.zeros(Cloud_Num)
            for i, cloud_time in enumerate(Cloud_Time):
                Mask = ((cloud_time-0.125) <= Time) & ((cloud_time+0.125) > Time)
            
                Diffuse_Fraction = Sd[Mask]/Sg[Mask]
                
                if np.nanmean(Diffuse_Fraction) >= 0.9:     #Overcast
                    Cloud_Type[i] = 2
                elif np.nanmean(Diffuse_Fraction) <= 0.3:     #Clear
                    Cloud_Type[i] = 1
                elif np.nanstd(Diffuse_Fraction) <= 0.05:     #Cumuliform
                    Cloud_Type[i] = 4
                elif np.nanstd(Diffuse_Fraction) >= 0.1:     #Stratiform
                    Cloud_Type[i] = 3
                else:                                    #Unclassified
                    Cloud_Type[i] = 5

            #If input time was in datetime, we now convert back to datetime
            if Datetime_Found == 'datetime': Cloud_Time = np.array([roundTime(Date + timedelta(hours=hour_frac), 1) for hour_frac in Cloud_Time], dtype=object)
            if Datetime_Found == 'datetime64': Cloud_Time = hours2dt(Cloud_Time, Date)
                
    return Cloud_Time, Cloud_Type
    
def degK2degC(Kelvin):
    """Converts either an array, list or value from kelvin to Celsius"""
    
    if isinstance(Kelvin, list):
        return (np.asarray(Kelvin, dtype=np.float64)-273.15).tolist()
    elif isinstance(Kelvin, np.ndarray):
        return Kelvin-273.15
    elif isinstance(Kelvin, float):
        return Kelvin-273.15
    else:
        sys.exit("[ERROR in degK2degC] Kelvin needs to be either a list, ndarray or float")
        
def radarbounds(Radar_Time, Radar_Height, Radar_Mag, Radar_TimeBounds, Lidar_TimeBounds, Cloud_Base, Lidar_Time):
    """Determines the cloud boundaries within a radar image. Requires all 3 dimensions of
    a radar image (e.g. time, height, magnitude), the index values denoting the boundaries
    of each cloud (in a 2D array) and the cloud base height array (1D) and must have the
    same dimensions as Radar_Time (i.e. Radar_Time.size == Cloud_Base.size).
    
    Parameters
    ----------
    Radar_Mag : ndarray or multiple ndarray
        Either a single ndarray of dimensions (height, time) or multiple ndarray with the same
        dimensions. E.g. Radar_Mag = (Radar, Radar) when calling radar-bounds. Each radar group
        with a tuple must have equal height and time dimensions."""
    
    #Create height mask to remove lowest band of clouds
    Radar_Height_Mask = (Radar_Height > 0.3).ravel()
    Radar_Height_Subset = Radar_Height[Radar_Height_Mask].ravel()
    
    #Select main radar product to detect clouds with. This is typically the radar reflectivity (Z).
    if isinstance(Radar_Mag, tuple):
        Radar_Mag_Main = Radar_Mag[0]
    else:
        Radar_Mag_Main = Radar_Mag
            
    Cloud_HeightBounds = zip(np.zeros(Radar_TimeBounds.shape[0]))
    for cloud_id, (radar_timebounds, lidar_timebounds) in enumerate(zip(Radar_TimeBounds, Lidar_TimeBounds)):
        
        #[Step 1] Subset Radar Reflectivity
        Cloud_Mag = Radar_Mag_Main[Radar_Height_Mask, radar_timebounds[0]:radar_timebounds[1]]
        
        #[Step 2] Determine number of nan's for each height layer
        #ClearSkies = np.sum(np.isnan(np.where(Cloud_Mag > -30,Cloud_Mag,np.nan)),axis=1)
        ClearSkies = np.sum(np.isnan(Cloud_Mag),axis=1)
        
        #[Step 3] Find lowest cloud base layer
        print(Lidar_Time[lidar_timebounds], np.sort(Cloud_Base[lidar_timebounds[0]:lidar_timebounds[1]]))
        CloudBaseMin = np.min(Cloud_Base[lidar_timebounds[0]:lidar_timebounds[1]])
        CloudBaseMax = np.max(Cloud_Base[lidar_timebounds[0]:lidar_timebounds[1]])
        CloudBase90     = np.nanpercentile(Cloud_Base[lidar_timebounds[0]:lidar_timebounds[1]], 10)
        
        #[Step 4] Find Lowest Cloud Top Layer
        mask = ClearSkies >= 0.98*np.diff(radar_timebounds)
        CloudTop = Radar_Height_Subset[mask]
        CloudTop = CloudTop[CloudTop > CloudBase90][0]
        
        #[Step 5] Remove all values below cloud base at each time-step
        Base = Cloud_Base[lidar_timebounds[0]:lidar_timebounds[1]]
        Cloud_Time = np.arange(radar_timebounds[0], radar_timebounds[1], dtype=int)
        Base_Mask = searchsorted(Radar_Height.ravel(), Base)
        for lower_limit, timestamp in zip(Base_Mask, Cloud_Time):
            if isinstance(Radar_Mag, tuple):
                for radar_type in xrange(len(Radar_Mag)):
                
                    try:
                        Radar_Mag[radar_type][:lower_limit, timestamp] = np.nan
                    except:
                        print("\nRadar_Mag[radar_type]", Radar_Mag[radar_type].shape)
                        print("lower_limit", lower_limit)
                        print("time-stamp", timestamp)
                        Radar_Mag[radar_type][:lower_limit, timestamp] = np.nan
            else:
                Radar_Mag[:lower_limit, timestamp] = np.nan
            
        #[Step 6] Remove all values above cloud top at each time-step
        Top_Mask = np.tile(searchsorted(Radar_Height.ravel(), CloudTop), Cloud_Time.size)
        for upper_limit, timestamp in zip(Top_Mask, Cloud_Time):
            if isinstance(Radar_Mag, tuple):
                for radar_type in xrange(len(Radar_Mag)):
                    Radar_Mag[radar_type][upper_limit:, timestamp] = np.nan
            else:
                Radar_Mag[upper_limit:, timestamp] = np.nan
            
        #Add cloud base and cloud top heights to array
        Cloud_HeightBounds[cloud_id] = [argnear(Radar_Height.ravel(), CloudBaseMin), 
            argnear(Radar_Height.ravel(),CloudBaseMax), 
            argnear(Radar_Height.ravel(),CloudTop)]
        
    Cloud_HeightBounds = np.array(Cloud_HeightBounds, dtype=int)

    if isinstance(Radar_Mag, tuple):
        output = (Cloud_HeightBounds,)
        for i in xrange(len(Radar_Mag)): output += (Radar_Mag[i],)
    else:
        output = (Cloud_HeightBounds, Radar_Mag)
    
    return output
    
def cloud_identifer_ascent(height, rh_ice, method='Zhang', verbose=False):
    """
    Calculates the cloud and moist layers within a radiosonde ascent 
    profile. Available methods are:
    
    1) Zhang et al. (2010)
    
    Parameters
    ----------
    
    Returns
    -------
    Cloud_ID : ndarray, dtype = np.int8
        An array with the same size of the input data (e.g. 'height')
        which contains the identifier for each cloud. Starting with
        1, for all height positions where a cloud was identified a 1
        will be used to identify the first cloud for all height 
        positions. Sequentially, the next cloud identified will be
        marked by a 2. N.B. The moist layer clouds are not identified
        within this array.
    Layer_Type : ndarray, dtype = np.int8
        An array with the same size of the input data (e.g. 'height')
        which contains the layer type at each height level (see notes
        for layer type classification). This is very similar to 
        Cloud_ID but does not differentiate between cloud layers.
    
    Notes
    -----
    Layer Type : Classification
        0 = Clear Air, 
        1 = Moist Layer, 
        2 = Cloud Layer.
            
    References
    ----------
    Zhang, J., H. Chen, Z. Li, X. Fan, L. Peng, Y. Yu, and M. Cribb 
        (2010). Analysis of cloud layer structure in Shouxian, China
        using RS92 radiosonde aided by 95 GHz cloud radar. J. 
        Geophys. Res., 115, D00K30, doi: 10.1029/2010JD014030.
    WMO, 2017. Clouds. In: Internal Cloud Atlas Manual on the 
        Observation of Clouds and Other Meteors. Hong Kong: WMO, 
        Section 2.2.1.2.
    """
    
    # Create Height-Resolving RH Thresholds (see Table 1 in 
    # Zhang et al. (2010)). N.B. use np.interp(val, 
    # RH_Thresholds['altitude'], RH_Thresholds['*RH']) where 
    # val is the height range you want the RH Threshold .
    RH_Thresholds = {'minRH' : [0.92, 0.90, 0.88, 0.75, 0.75],
        'maxRH' : [0.95, 0.93, 0.90, 0.80, 0.80],
        'interRH' : [0.84, 0.82, 0.78, 0.70, 0.70],
        'altitude' : [0, 2, 6, 12, 20]}

    # Define the cloud height levels (km) as defined by WMO (2017).
    Z_Levels = {'low' : [0,2], 'middle' : [2,7], 'high' : [5,13]}

    # Define the types of layers that can be detected.
    Cloud_Types = {0 : 'Clear Air', 1 : 'Moist (Not Cloud)', 2 : 'Cloud'}

    # Define the min, max and interRH for all measure altitudes
    minRH = np.interp(height, 
                      RH_Thresholds['altitude'], 
                      RH_Thresholds['minRH'], 
                      left=np.nan, 
                      right=np.nan)*100
    maxRH = np.interp(height, 
                      RH_Thresholds['altitude'], 
                      RH_Thresholds['maxRH'], 
                      left=np.nan, 
                      right=np.nan)*100
    interRH = np.interp(height, 
                        RH_Thresholds['altitude'], 
                        RH_Thresholds['interRH'], 
                        left=np.nan, 
                        right=np.nan)*100

    # [Step 1]: The base of the lowest moist layer is determined
    # as the level when RH exceeds the min-RH corresponding to
    # this level. N.B. Catch simple warnings required because of
    # np.nan values in either RH or minRH arrays. Can't remove
    # them due to mask not having correct size and masking the
    # invalid values still throws a warning message.
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")

        minRH_mask = (rh_ice > minRH)

    # [Step 2 and 3]: Above the base of the moist layer, 
    # contiguous levels with RH over the corresponding min-RH 
    # are treated as the same layer.
    height[~minRH_mask] = np.nan
    Clouds_ID = contiguous(height, 1)

    # [Step 4]: Moist layers with bases lower than 120m and 
    # thickness's less than 400m are discarded.
    for Cloud in np.unique(Clouds_ID)[1:]:
        if height[Clouds_ID == Cloud][0] < 0.12:
            if height[Clouds_ID == Cloud][-1] - height[Clouds_ID == Cloud][0] < 0.4:
                Clouds_ID[Clouds_ID == Cloud] = 0

    # [Step 5]: The moist layer is classified as a cloud layer 
    # is the maximum RH within this layer is greater than the 
    # corresponding max-RH at the base of this moist layer.
    LayerType = np.zeros(height.size, dtype=int) # 0: Clear Air, 1: Moist Layer, 2: Cloud Layer
    for Cloud in np.unique(Clouds_ID)[1:]:
        if np.any(rh_ice[Clouds_ID == Cloud] > maxRH[Clouds_ID == Cloud][0]):
            LayerType[Clouds_ID == Cloud] = 2
        else:
            LayerType[Clouds_ID == Cloud] = 1

    # [Step 6]: The base of the cloud layers is set to 280m AGL, 
    # and cloud layers are discarded if their tops are lower 
    # than 280m    
    for Cloud in np.unique(Clouds_ID)[1:]:
        if height[Clouds_ID == Cloud][-1] < 0.280:
            Clouds_ID[Clouds_ID == Cloud] = 0
            LayerType[Clouds_ID == Cloud] = 0

    # [Step 7]: Two contiguous layers are considered as one-
    # layer cloud if the distance between these two layers is
    # less than 300m or the minimum RH within this distance is
    # more than the maximum inter-RG value within this distance
    for Cloud_Below, Cloud_Above in zip(np.unique(Clouds_ID)[1:-1], np.unique(Clouds_ID)[2:]):
        
        #Define the index between clouds of interest
        Air_Between = np.arange(bool2int(Clouds_ID == Cloud_Below)[-1], bool2int(Clouds_ID == Cloud_Above)[0])
        
        if ((height[Clouds_ID == Cloud_Above][0] - height[Clouds_ID == Cloud_Below][-1]) < 0.3) or (np.nanmin(rh_ice[Air_Between]) > np.nanmax(interRH[Air_Between])):
            Joined_Cloud_Mask = np.arange(bool2int(Clouds_ID == Cloud_Below)[0], bool2int(Clouds_ID == Cloud_Above)[-1])
            
            #Update the cloud ID array as the Cloud_Below and Cloud_Above are not distinct clouds
            Clouds_ID[Joined_Cloud_Mask] = Cloud_Below
            
            #Update the LayerType to reflect the new cloud merging
            if np.any(LayerType[Clouds_ID == Cloud_Below] == 2) or np.any(LayerType[Clouds_ID == Cloud_Above] == 2):
                LayerType[Joined_Cloud_Mask] = 2
            else:
                LayerType[Joined_Cloud_Mask] = 1
        
    # [Step 8] Clouds are discarded if their thickness's are less than 30.5m for low clouds and 61m for middle/high clouds
    for Cloud in np.unique(Clouds_ID)[1:]:
        if height[Clouds_ID == Cloud][0] < Z_Levels['low'][1]:
            if height[Clouds_ID == Cloud][-1] - height[Clouds_ID == Cloud][0] < 0.0305:
                Clouds_ID[Clouds_ID == Cloud] = 0
                LayerType[Clouds_ID == Cloud] = 0

        else:
            if height[Clouds_ID == Cloud][-1] - height[Clouds_ID == Cloud][0] < 0.0610:
                Clouds_ID[Clouds_ID == Cloud] = 0
                LayerType[Clouds_ID == Cloud] = 0
    
    #Re-update numbering of each cloud identified
    Clouds_ID = contiguous(Clouds_ID, invalid=0)
    
    #Output verbose to screen
    if verbose is True:
        print("Detected Clouds and Moist Layers\n--------------------------------")
        for Cloud in np.unique(Clouds_ID)[1:]:
            print("Cloud %s. Cloud Base = %.2fkm, Cloud Top = %.2fkm, Layer Type: %s" % (Cloud, height[Clouds_ID == Cloud][0], height[Clouds_ID == Cloud][-1], Cloud_Types[LayerType[Clouds_ID == Cloud][0]]))
    
    return Clouds_ID, LayerType
    


from __future__ import absolute_import, division, print_function
import sys
import numpy as np
import itertools, warnings
import time as systime

from .manipulation import argnear

__doc__ = "Built for Python 2.7. Upgrading to Python 3.7 soon!"

# Compatability for python3
if getattr(sys.version_info, 'major') == 3:
	xrange = range
	
def argpercentile(x, percentile):
    """Outputs the index of the percentile value for the data.
    
    x is 1D
    percentile is between 0 and 100
    
    Reference
    ---------
    https://stackoverflow.com/a/26071170"""
    
    idx = int(np.round(percentile/100 * (len(x) - 1))); 
    
    return np.argpartition(x,idx)[idx]

def cosarctan(x):
    """Calculates the trigonometric function: Cos(Arctan(x)) by simplifying
    to (x^2+1)^-0.5. This function is meant to provide an improvement in 
    speed over long iterations"""
    
    return (x**2+1)**-0.5
    
def power(array, pow):
    """Calculates the power of a function that can handle negative numbers
    in array. The method removes the sign of each element in array, calculates
    the power like normal and then returns the sign afterwards. This is a 
    workaround for C language issues"""

    #Convert to numpy array if not already
    array = np.asarray(array)
    
    #Get signs of all elements in the array
    sign = np.sign(array)
    
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
    
        return sign*(np.abs(array)**pow)

def interavg(arr, n, type='mean'):
    """Sums over n elements in an array.
    
    Reference
    ---------
    https://stackoverflow.com/a/15956341/8765762
    """
    
    return getattr(np, type)(arr.reshape(-1, n), axis=1)
    
def isprime(a):
    return all(a % i for i in xrange(2, a))

def geomspace(start, stop, num=50, endpoint=True, dtype=None):
    """Creates an array between start and stop using a log10
    spacing grid. This function is meant to copy np.geomspace
    when using numpy before 1.12"""
    
    array = np.logspace(np.log10(start) , np.log10(stop), num=num)
    
    return array.astype(dtype)
    
def get_multipler(array, m=2, mode='1D'):
    """Determines multiplier at each step which produces a whole
    number.
    
    Parameters
    ----------
    array : int
        The number you want to divide by
    m : int, optional
        Initial division number. By default this is set to 2
        as this is the lowest integer divisor
        
    Returns
    -------
    mul : numpy array
        The array sequence of multipliers at each step. This will be
        in reversed order so if we start at the lowest number (i.e.
        first number in mul) then we can multiply by the first number 
        in mul and so on
    
    """
    
    #if number is prime then remove 1 from sequence
    if isprime(array) == True: 
        array += 1
        #print("ITS THE PRIME OF YOUR LIFE!")
    
    #if mode == '1D':
    mul = np.array([], dtype=int)
    while array > m:
        array_temp = array/m
        if array_temp == int(array_temp):
            array//=m
            mul = np.append(mul, m)
        else:
            while array_temp != int(array_temp):
                if m < 10:
                    m += 1
                    array_temp = array/m
                else:
                    array += 1
                    m = 2
                    array_temp = array/m
            array//=m
            mul = np.append(mul, m)
            
    mul = np.append(mul, m)

    mul = np.array(list(reversed(mul)), dtype=int)
    
    return mul
    
def complex2float(array):
    """Sums real and imaginary parts of an array together element-
    wise. This is used for finding the power of negative numbers
    which are calcuable but not in numpy which returns a nan.
    
    This is a workaround routine and is typically using as follows:
    
    complex2float(np.array(array, dtype=complex)**power)
    
    """
    
    return array.real + array.imag

def disjoint_permutations_v2(subsets):        
    """Next-gen of disjoint permutations
    
    This seems to be much quicker than the other permutation method. 
    A downside though is that the input subset needs to be an M x N 
    matrix rather than having subsubsets of different lengths.
    """
    
    
    t0 = systime.time()
    
    #Initial conditions
    perms = []
    remi = lambda a: a != -1
    flatten = lambda l: [item for sublist in l for item in sublist]
    subsets_length = [range(len(subsets[i])) for i in xrange(len(subsets))]
    
    #Pad out subsets input to detect all combinations. Otherwise we will only detect combinations of length subset columns
    subsets_column = np.array(subsets).T
    subsets_column = np.lib.pad(subsets_column, [(len(subsets_column)-1, len(subsets_column)-1), (0,0)], 'constant', constant_values=-1)
    
    t1 = systime.time()
    #Create Iteration
    #subsets_iters = np.column_stack(([np.tile(np.repeat(subsets_length[i], len(subsets)**(len(subsets)-1-i)),  len(subsets)**i) for i in xrange(len(subsets))]))
    subsets_iters = np.column_stack(([np.tile(np.repeat(subsets_length[i], len(subsets[i])**(len(subsets)-1-i)),  len(subsets[i])**i) for i in xrange(len(subsets))]))
    t2 = systime.time()
    
    subsets_column_temp = subsets_column.copy()
    lensubsets = len(subsets_column)
    for i in subsets_iters: #iterate over lock formation (i.e. [1,1,1],[1,1,2],[1,1,3],[1,2,1] etc.)
        for j in xrange(len(subsets_iters[i])):
            subsets_column_temp[:,j] = np.roll(subsets_column[:,j], i[j], axis=0)
            for k in xrange(lensubsets):
                perms.append(tuple(subsets_column_temp[k].copy().tolist()))

    t3 = systime.time()    
        
    #Clean up: Remove -1 from all sublists
    perms = [filter(remi, x) for x in perms]
    
    t4 = systime.time()
    #Remove () tuples
    perms = [x for x in perms if x]

    t5 = systime.time()
    #Remove doubles
    perms = np.sort([x for x in set(x for x in perms)])

    print("Time: %.5f, %.5f, %.5f, %.5f, %.5f, %.5f, %.5f, Length: %.0f" % (t1-t0, t2-t1, t3-t2, t4-t3, t5-t4, systime.time()-t5, systime.time()-t0, len(perms)))
    return perms

def disjoint_permutations(subsets, andxor=False):    
    ############################################################################
    """Determine Permutations of the Distributions.
    
    Method
    ------
    For a structured sequence where each subset is disjoint from one another we
    can use a cycle decomposition type approach.
    
    We determine the intersections between 2 cases of permutations:
    (1) Permutations of the subsets combined, e.g.:
            a = [1,2,3,4,5,6,7,8,9]
    (2) Permutations of the disjoint subsets, e.g.:
            b = [[1,2,3,4],[5,6,7],[7,8]]
    
    At the end we need to determine the common permutations in both datasets
    (i.e. A ^ B). This will lead to us getting all the correct combinations.
    
    This method can also be used to find the permutations with a subsets that have
    disjoint and union sets, e.g.:
            
            b = [[1,2,3,4],[5,6,7],[8,9],[8,9]]
            
    This basically overcomes the issue that we can't permutate within subsets when
    there are special cases which need to be self permutating.
    
    (see http://math.stackexchange.com/questions/1854141/permutation-of-disjoint-sets-of-a-symmetric-group
    for for information)
    
    Parameters
    ----------
    subsets : list
        A series of subsets that which have union with the set. Each subset can
        be disjoint or union with one another.
        
    andxor : boolean, optional
        Used when you want to satisfy the condition of input elements being
        completely distinct from each other (e.g. if a and b are distributions
        then,
            
                                set(a)^(set(a)&set(b)),
                                
        would be the elements in a that don't occur in b.
           
        Basically this just doubles the size of the perms array
    Returns
    -------
    perms : list
        A series of permutations 
    
    """
    ############################################################################
    
    t0 = systime.time()
    
    flatten = lambda l: [item for sublist in l for item in sublist]
    subsets_flat = np.unique(flatten(subsets)).tolist()

    #Get all permutations with repeats
    res = []
    #print("np.array(subsets).shape[0]",np.array(subsets).shape[0])
    for i in xrange(np.array(subsets).shape[0]):
        for j in xrange(np.array(subsets).shape[0]):
            for k in xrange(np.array(subsets).shape[0]-1):
                res.append(itertools.product(*subsets[j:j+1+i]))
    
    #Get all permutations for all structures
    res2 = []
    for l in range(1, len(subsets_flat)+1):
        for x in itertools.combinations(subsets_flat, l):
            res2.append(x)

    #Find common elements
    if andxor == True: 
        perms = sorted(list(set(flatten(res))|set(res2)))
    else:
        perms = sorted(list(set(flatten(res))&set(res2)))
    
    print("Time: %.2f, Length: %.0f" % (systime.time()-t0, len(perms)))
    return perms



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="gilly-utilities",
    version="0.9.6",
    author="James Gilmore",
    author_email="jamesgillygilmore@gmail.com",
    description="A data analysis package built upon Numpy, SciPy and Pandas",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/pacificgilly1992/Gilly_Utilities",
    packages=setuptools.find_packages(),
    install_requires=[
        'numpy',
        'scipy!=1.3.0',
        'pandas',
        'matplotlib',
        'statsmodels',
        'scikit-learn',
        'scikits.bootstrap',
        'patsy'
    ],
    classifiers=[
        "Programming Language :: Python :: 2",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: OS Independent",
    ],
)


import sys
import warnings 

import numpy as np
from datetime import datetime

from .system import iszero, isval, isarray
from .manipulation import array, flatten
from .datetime64 import isnat, dt2hours, hours2dt

# Compatability for python3
if getattr(sys.version_info, 'major') == 3:
    xrange = range

def _simple_slice(arr, inds, axis):
    """
    This does the same as np.take() except only supports simple slicing, 
    not advanced indexing, and thus is much faster.
    
    References
    ----------
    https://stackoverflow.com/a/37729566
    """

    sl = [slice(None)] * arr.ndim
    sl[axis] = inds

    return arr[tuple(sl)]

def _antiany(arr, func, axis=1, how='any', unpack=False, sortby=False, return_mask=False):
    """
    Worker function to remove any value specified using func along a
    specific axis.
    
    Parameters
    ----------
    
    arr : tuple of 1D array_like or non-tuple array_like
        If arr is a tuple, it is assumed that each element is a 
        column and will be transposed before any processing is 
        performed. Otherwise arr is treated "as is" and will be 
        processed directly.
    func : module
        A module definition used to specify the type of values to 
        remove. Examples are func=np.isnan, func=np.isinf. Be careful 
        on specifying the polarity for these functions. For example 
        func=np.isnan will remove all nan values despite np.isnan will
        return True for any values that are nan. Issue might occur if 
        you specify np.isfinite, which would remove all finite values, 
        despite the usual use would be to remove both np.nan and np.inf 
        values simultaneously. In this specific case you want to 
        specify func=lambda x: ~np.isfinite(x).
    axis : int, optional
        The axis you want to remove an np.nan values from. Here is a 
        help guide to picking the correct axis values:
            
            axis = 0 : column wise
            axis = 1 : row wise (DEFAULT)
    how : str, optional
        Test whether the entire axis contains an invalid value or whether
        only 1 invalid value allows removal. Only works on 2D data and
        uses the axis parameter. Default is 'any'.
    unpack : boolean, optional
        Determines whether to unpack the array into individual columns,
        useful if you originally had N column arrays which you want to 
        temporarily merge for invalid values.
    sortby : int, default is False
        Specify if you want to sort the data by a particular column. 
        Default is false and does not sort data, but other options are 
        to specify the column which you want to sort by.
    return_mask : bool, optionl, default is False
        Specify if you want to return the mask of invalid values along 
        the correct axis you specified, instead of the corrected arr.
        
    Returns
    -------
    
    arr_nonan : ndarry of floats or list of arrays dependent upon andxor
        The finalised ndarray with all nans removed will be returned 
        unless andxor is False where a list of column arrays will be 
        returned.
    
    Examples
    --------
    
    >>> x = np.array([1,2,np.nan,4,5])
    >>> y = np.array([20,30,40,np.nan,np.nan])
    
    >>> xy_new = antinan(np.array([x,y])
    
    >>> xy_new
    array([[1., 2.],
           [20., 30.]])

    References
    -------
    http://stackoverflow.com/questions/27532503/remove-nan-row-from-x-array-and-also-the-corresponding-row-in-y
    https://www.w3resource.com/python-exercises/numpy/python-numpy-exercise-91.php
    https://stackoverflow.com/a/37729566
    
    """

    # Force arr to be a numpy array
    if isinstance(arr, tuple):
        arr = np.asarray(arr).T
    else:   
        arr = np.asarray(arr)
    
    # Check if arr has any data
    if arr.size == 0:
        return np.array([])
    
    # For 1D arrays
    if arr.ndim == 1:

        # Calculate mask of invalid values
        mask = ~func(arr)

        # Return mask if necessary
        if return_mask:
            return mask

        # Sort data if necessary 
        if sortby is not False:
            return np.sort(arr[mask], kind='mergesort')
        else:
            return arr[mask]

    # For ND arrays
    elif arr.ndim > 0:

        # Determine index for all nan values in each column
        if how == 'any':
            mask = ~func(arr).any(axis=axis)
        elif how == 'all':
            mask = ~func(arr).all(axis=axis)
        else:
            raise ValueError("how parameter can only be either 'all' or 'any'")

        # Return mask if necessary
        if return_mask:
            return mask

        # Remove all rows with a nan
        arr_noany = _simple_slice(arr.T, mask, axis=axis).T

    else:
        return arr

    # Sort array by column if necessary
    if sortby is not False: 
        arr_noany = arr_noany[arr_noany[:,int(sortby)].argsort(kind='mergesort')]

    if unpack:
        return tuple(arr_noany.reshape(-1, arr_noany.shape[1]).T)
    else:
        return arr_noany

def antifinite(arr, axis=1, how='any', andxor=True, unpack=False, sortby=False, 
        return_mask=False, oldmethod=False):
    """
    Removes any np.inf and np.nan values from arr along axis.
    
    See _antiany for __doc__
    
    oldmethod : bool, optional
        Specifiy to use the old method for removing nan and inf values 
        (used for backwards compatability, or False to use new method
    
    Update (For oldmethod)
    ------
    Added support for python datetime and numpy datetime64 objects automatically. This
    is achieved by converting any column with datetimes into floating point numbers.
    Run antifinite as normal and then convert back to datetime objects again automatically.
    """

    if oldmethod:
        #For 1D arrays
        if len(np.shape(arr)) == 1 and np.array(arr).dtype != 'O':
            
            #Check array does not contain datetime function. If so convert to number temporarily
            arr, check, dtype = dt2hours(arr, epoch=np.datetime64("1900-01-01"), checkback=True)

            if sortby is not False:
                if check is False:
                    return np.sort(arr[np.isfinite(arr)], kind='mergesort')
                else: 
                    return hours2dt(np.sort(arr[np.isfinite(arr)], kind='mergesort'), epoch=np.datetime64("1900-01-01"), dtype=dtype)
            else:
                if check is False:
                    return arr[np.isfinite(arr)]  
                else:
                    return hours2dt(arr[np.isfinite(arr)], epoch=np.datetime64("1900-01-01 00:00:00"), dtype=dtype)

        #For ND arrays
        if np.shape(arr)[0] > 0:
            
            #Convert datetime columns to hours
            array_has_datetime     = np.zeros(np.shape(arr)[0], dtype=bool)
            array_dtype            = np.zeros(np.shape(arr)[0], dtype='S25')
            for i in xrange(np.shape(arr)[0]): 
                if isinstance(arr[i][0], np.datetime64) or isinstance(arr[i][0], datetime):
                    array_has_datetime[i] = True
                    array_dtype[i] = arr[i][0].dtype
                else:
                    array_has_datetime[i] = False
                    array_dtype[i] = arr[i][0].dtype
            arr = np.asarray(arr, dtype=float)    

            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                arr[arr <= -10e+17] = np.nan

            if andxor == True:
                #Determine index for all nan values in each column
                x_mask = np.any(np.isfinite([arr[0]]), axis=0)
                
                for i in xrange(1, arr.shape[0]): x_mask &= np.any(np.isfinite([arr[i]]), axis=0)
                
                #Remove all rows with a nan
                x_nonan = arr.T[x_mask]

            else:
                #Split x into column arrays
                x_arrays      = zip(np.zeros([arr.shape[0]]))
                x_nonan = zip(np.zeros([arr.shape[0]]))
                for i in xrange(arr.shape[0]): x_arrays[i] = arr[i]
                
                #Remove nans from each column individually
                for i in xrange(arr.shape[0]): x_nonan[i] = x_arrays[i][np.any(np.isfinite([x_arrays[i]]), axis=0)]
        else:
            #return np.split(arr.T, arr.shape[0])
            return arr

        #Sort array by column if specified
        if sortby is not False: x_nonan = x_nonan[x_nonan[:,int(sortby)].argsort(kind='mergesort')]

        #Return original dtype
        out_data = zip(np.zeros(len(arr)))
        for i in xrange(len(arr)):
            out_data[i] = x_nonan[:,i].astype(np.float64).astype(array_dtype[i])

        if unpack:
            return out_data    
        else:
            return np.vstack(out_data)

        # #Convert arrays that contained datetimes
        # if not np.all(array_dtype == array_dtype[0]): x_nonan = x_nonan.astype(object)
        # for i in xrange(arr.shape[0]): x_nonan[:,i] = x_nonan[:,i].astype(np.float64).astype(array_dtype[i])

        # x_nonan[:,1] = x_nonan[:,1].astype(np.float64)
        # print("x_nonan[:,1]", x_nonan[:,1].dtype)
        # if unpack == True:
            # return flatten(np.split(x_nonan.T, arr.shape[0]))
        # else:

            # return x_nonan

    else:
        return _antiany(arr, lambda x: ~np.isfinite(x), axis=axis,
                how=how, unpack=unpack, sortby=sortby, 
                return_mask=return_mask)

def antinan(arr, axis=1, how='any', unpack=False, sortby=False, 
        return_mask=False):
    """
    Removes any np.nan values from arr along axis.
    
    See _antiany for __doc__
    """

    return _antiany(arr, np.isnan, axis=axis, how=how, unpack=unpack, 
            sortby=sortby, return_mask=return_mask)

def antiinf(arr, axis=1, how='any', unpack=False, sortby=False, 
        return_mask=False):
    """
    Removes any np.inf values from arr along axis.
    
    See _antiany for __doc__
    """

    return _antiany(arr, np.isinf, axis=axis, how=how, unpack=unpack, 
            sortby=sortby, return_mask=return_mask)

def antinat(arr, axis=1, how='any', unpack=False, sortby=False, 
        return_mask=False):
    """
    Removes any NaT datetimes from arr along axis.
    
    See _antiany for __doc__
    """

    return _antiany(arr, isnat, axis=axis, how=how, unpack=unpack, 
            sortby=sortby, return_mask=return_mask)      

def antizero(arr, axis=1, how='any', unpack=False, sortby=False, 
        return_mask=False):
    """
    Removes any zero values from arr along axis.
    
    See _antiany for __doc__
    """

    return _antiany(arr, iszero, axis=axis, how=how, unpack=unpack, 
            sortby=sortby, return_mask=return_mask)      

def antival(arr, val, axis=1, how='any', unpack=False, sortby=False, 
        return_mask=False):
    """
    Removes any values from arr along axis.
    
    See _antiany for __doc__
    """

    return _antiany(arr, lambda x: isval(x, val), axis=axis, how=how, 
            unpack=unpack, sortby=sortby, return_mask=return_mask)      

def conditional(arr, conditions, operators=None, verbose=False):
    """
    Returns a boolean array of the conditions on arr, taking into account
    invalid values.
    
    conditions needs to be an array_like containing string of conditions
    
    e.g. conditions = ("> 2", "< 10")
    
    operators needs to be an array_like containing string of operations
    
    e.g. operators = ("&",)
    
    with length 1 less than conditions.
    
    Overall this follows a conditions as so

    Reference
    ---------
    https://stackoverflow.com/a/25346972
    """
    
    if not isarray(arr):
        raise ValueError("arr needs to be array_like")
    if len(conditions) > 1 and operators is None:
        raise ValueError("need to specifiy operators if more than 1 conditions is specified")
    if len(conditions) - len(operators) != 1:
        raise ValueError("operators needs to have a length 1 less than conditions")
    
    # Convert arr to numpy if required
    arr = np.atleast_1d(arr)
    conditions = np.atleast_1d(conditions)
    operators = np.atleast_1d(operators)
    
    if (conditions.ndim > 1) or (operators.ndim > 1): 
        raise ValueError("conditions and operators must be 1 dimensional")
    
    # Determine invalid values
    mask = np.isfinite(arr)
    
    # Loop through each conditional
    if conditions.size > 1:
        full_conditions = ''
        for cond, op in zip(conditions, operators):
            full_conditions += '(arr[mask] ' + cond + ') ' + op + ' '
            
        full_conditions += '(arr[mask] ' + conditions[-1] + ')'
    else:
        full_conditions = 'arr[mask] ' + cond
    
    if verbose:
        print("full_conditions", full_conditions)
    
    # Perform conditions
    mask[mask] &= eval(full_conditions)
    
    # Return mask
    return mask
    
def nan_helper(y):
    """
    Helper to handle indices and logical indices of NaNs.

    Input:
        - y, 1d numpy array with possible NaNs
    Output:
        - nans, logical indices of NaNs
        - index, a function, with signature indices= index(logical_indices),
          to convert logical indices of NaNs to 'equivalent' indices
    Example:
        >>> # linear interpolation of NaNs
        >>> nans, x= nan_helper(y)
        >>> y[nans]= np.interp(x(nans), x(~nans), y[~nans])
    
    References
    ----------
    https://stackoverflow.com/a/6520696
    """

    return np.isnan(y), lambda z: z.nonzero()[0]    


from __future__ import absolute_import
import sys
import numpy as np
from math import cos,sin,acos,asin,tan  
from math import degrees as deg, radians as rad  
from datetime import date,datetime,time  
  
# Compatability for python3
if getattr(sys.version_info, 'major') == 3:
	xrange = range
	
class Earth:  
	"""  
	Calculate sunrise and sunset based on equations from NOAA 
	http://www.srrb.noaa.gov/highlights/sunrise/calcdetails.html 

	typical use, calculating the sunrise at the present day: 

	import datetime 
	import sunrise 
	s = sun(lat=49,long=3) 
	print('sunrise at ',s.sunrise(when=datetime.datetime.now()) 
	
	Reference
	---------
	https://michelanders.blogspot.com/2010/12/calulating-sunrise-and-sunset-in-python.html
	
	N.B. This class function is not perfect. Other methods are better
	at calculating the sunrise and sunset times (e.g. PyEphem)
	
	See https://rhodesmill.org/pyephem/rise-set.html#computing-twilight
	
	"""  
	def __init__(self,lat=51.44142,long=-0.93776): # default Amsterdam  
		self.lat=lat  
		self.long=long  

	def sunrise(self,when=None):  
		""" 
		return the time of sunrise as a datetime.time object 
		when is a datetime.datetime object. If none is given 
		a local time zone is assumed (including daylight saving 
		if present) 
		"""  
		if when is None : when = datetime.now(tz=LocalTimezone())  
		self.__preptime(when)  
		self.__calc()  
		return self.__timefromdecimalday(self.sunrise_t)  

	def sunset(self,when=None):  
		if when is None : when = datetime.now(tz=LocalTimezone())  
		self.__preptime(when)  
		self.__calc()  
		return self.__timefromdecimalday(self.sunset_t)  

	def solarnoon(self,when=None):  
		if when is None : when = datetime.now(tz=LocalTimezone())  
		self.__preptime(when)  
		self.__calc()  
		return self.__timefromdecimalday(self.solarnoon_t)  

	@staticmethod  
	def __timefromdecimalday(day):  
		""" 
		returns a datetime.time object. 

		day is a decimal day between 0.0 and 1.0, e.g. noon = 0.5 
		"""  
		hours  = 24.0*day  
		h      = int(hours)  
		minutes= (hours-h)*60  
		m      = int(minutes)  
		seconds= (minutes-m)*60  
		s      = int(seconds)  
		return time(hour=h,minute=m,second=s)  

	def __preptime(self,when):  
		""" 
		Extract information in a suitable format from when,  
		a datetime.datetime object. 
		"""  
		# datetime days are numbered in the Gregorian calendar  
		# while the calculations from NOAA are distibuted as  
		# OpenOffice spreadsheets with days numbered from  
		# 1/1/1900. The difference are those numbers taken for   
		# 18/12/2010  
		self.day = when.toordinal()-(734124-40529)  
		t=when.time()  
		self.time= (t.hour + t.minute/60.0 + t.second/3600.0)/24.0  

		self.timezone=0  
		offset=when.utcoffset()  
		if not offset is None:  
			self.timezone=offset.seconds/3600.0  

	def __calc(self):  
		""" 
		Perform the actual calculations for sunrise, sunset and 
		a number of related quantities. 

		The results are stored in the instance variables 
		sunrise_t, sunset_t and solarnoon_t 
		"""  
		timezone = self.timezone # in hours, east is positive  
		longitude= self.long     # in decimal degrees, east is positive  
		latitude = self.lat      # in decimal degrees, north is positive  

		time  = self.time # percentage past midnight, i.e. noon  is 0.5  
		day      = self.day     # daynumber 1=1/1/1900  

		Jday     =day+2415018.5+time-timezone/24 # Julian day  
		Jcent    =(Jday-2451545)/36525    # Julian century  

		Manom    = 357.52911+Jcent*(35999.05029-0.0001537*Jcent)  
		Mlong    = 280.46646+Jcent*(36000.76983+Jcent*0.0003032)%360  
		Eccent   = 0.016708634-Jcent*(0.000042037+0.0001537*Jcent)  
		Mobliq   = 23+(26+((21.448-Jcent*(46.815+Jcent*(0.00059-Jcent*0.001813))))/60)/60  
		obliq    = Mobliq+0.00256*cos(rad(125.04-1934.136*Jcent))  
		vary     = tan(rad(obliq/2))*tan(rad(obliq/2))  
		Seqcent  = sin(rad(Manom))*(1.914602-Jcent*(0.004817+0.000014*Jcent))+sin(rad(2*Manom))*(0.019993-0.000101*Jcent)+sin(rad(3*Manom))*0.000289  
		Struelong= Mlong+Seqcent  
		Sapplong = Struelong-0.00569-0.00478*sin(rad(125.04-1934.136*Jcent))  
		declination = deg(asin(sin(rad(obliq))*sin(rad(Sapplong))))  

		eqtime   = 4*deg(vary*sin(2*rad(Mlong))-2*Eccent*sin(rad(Manom))+4*Eccent*vary*sin(rad(Manom))*cos(2*rad(Mlong))-0.5*vary*vary*sin(4*rad(Mlong))-1.25*Eccent*Eccent*sin(2*rad(Manom)))  

		hourangle= deg(acos(cos(rad(90.833))/(cos(rad(latitude))*cos(rad(declination)))-tan(rad(latitude))*tan(rad(declination))))  

		self.solarnoon_t=(720-4*longitude-eqtime+timezone*60)/1440  
		self.sunrise_t  =self.solarnoon_t-hourangle*4/1440  
		self.sunset_t   =self.solarnoon_t+hourangle*4/1440  
		
def haversine(latlong1, latlong2, units='rad'):
	"""
	Calculate the great circle distance between two points 
	on the earth (specified in decimal degrees) and the
	angle between them
    
	Ref: https://stackoverflow.com/a/4913653/8765762
	Ref: https://stackoverflow.com/a/15424841/8765762
	"""
	
	#Constants
	r = 6371000 # Radius of earth in metres. Use 3956 for miles
	
    # convert decimal degrees to radians 
	lat1, lon1, lat2, lon2 = map(np.radians, latlong1+latlong2)

    # Haversine formula 
	dlon = lon2 - lon1 
	dlat = lat2 - lat1 
	a = np.sin(dlat/2)**2 + np.cos(lat1) * np.cos(lat2) * np.sin(dlon/2)**2
	d = 2 * r * np.arcsin(np.sqrt(a)) 
	
	#Calculate angle between two latlong coords
	theta = np.pi+np.arctan2(np.sin(lon2-lon1)*np.cos(lat2), np.cos(lat1)*np.sin(lat2) - np.sin(lat1)*np.cos(lat2)*np.cos(lon2-lon1))
	
	if units == 'deg': theta *= 180/np.pi
	
	#Return distance in metres between latlong1 and latlong2 and the central angle between them
	return d, theta	